#include "squid.h"
#include <errno.h>

/* Semantics:  If connID == 0, we don't print anything, b/c it's
 * a connection for which we have no data.
 */

static Logfile *logfile = NULL;
static int LogfileStatus = LOG_DISABLE;
static u_int32_t connID = 0;
static u_int32_t connSubID = 1;

conn_id_t connID_none = { 0, 0 };

#define LOG_PREAMBLE() do { \
    getCurrentTime(); \
} while (0);

void
connLogRotate(void)
{
    if (NULL == logfile)
	return;
    logfileRotate(logfile);
}

void
connLogInit(void)
{
    if (!Config.Log.conn || 
	(strcasecmp(Config.Log.conn, "none") == 0))
	return;
    logfile = logfileOpen(Config.Log.conn, 0, 1);
    LogfileStatus = LOG_ENABLE;
}

void
connLogClose(void)
{
    if (logfile) logfileClose(logfile);
    logfile = NULL;
}

/*
 * Log a request iff it's a cache miss and will require
 * subsequent processing.  Assumes the client is allowed
 * to make the request, it's a TCP query, etc.
 * (in short, it will require a connection. ;-)
 */

static conn_id_t
connLogNewReq()
{
    conn_id_t cid;
    
    connID++;

    cid.connID = connID;
    cid.connSubID = 0;
    return cid;
}

extern void
connLogNewSub(conn_id_t *cid)
{
    connSubID++;
    cid->connSubID = connSubID;
}

conn_id_t
connLogNewClientReq(const char *url, int from_cache, struct in_addr client)
{
    conn_id_t reqid = connLogNewReq();
    LOG_PREAMBLE();
    if (LogfileStatus == LOG_ENABLE)
	logfilePrintf(logfile, "%ld.%6.6ld %u new %d %s %s\n",
		      current_time.tv_sec, current_time.tv_usec, reqid.connID,
		      from_cache ? 1 : 0,
		      inet_ntoa(client), url);
    return reqid;
}

conn_id_t
connLogNewICPReq(const char *url, struct in_addr srca, u_short srcp,
		 struct in_addr dsta, u_short dstp, u_int32_t icp_src,
		 u_int32_t icp_reqnum, int do_setup)
{
    char srcdude[64];
    char dstdude[64];
    conn_id_t reqid = connLogNewReq();
    
    if (LogfileStatus != LOG_ENABLE) return connID_none;

    LOG_PREAMBLE();
    
    strcpy(srcdude, inet_ntoa(srca));
    strcpy(dstdude, inet_ntoa(dsta));

    logfilePrintf(logfile,
		  "%ld.%6.6ld %u newicp %u %u %s:%d %s:%d %d %s\n",
		  current_time.tv_sec, current_time.tv_usec, reqid.connID,
		  icp_src, icp_reqnum,
		  srcdude, ntohs(srcp),
		  dstdude, ntohs(dstp),
		  do_setup,
		  url);
    return reqid;
}

void
connLogConnStart(conn_id_t reqid, struct in_addr srca, u_short srcp,
		 struct in_addr dsta, u_short dstp, int fd)
{
    char srcdude[64];
    char dstdude[64];

    strcpy(srcdude, inet_ntoa(srca));
    strcpy(dstdude, inet_ntoa(dsta));
    
    if (LogfileStatus != LOG_ENABLE) return;
    if (reqid.connID == 0) return;
    LOG_PREAMBLE();
    logfilePrintf(logfile,
		  "%ld.%6.6ld %lu connstart %lu %s:%d %s:%d %d\n",
		  current_time.tv_sec, current_time.tv_usec, reqid.connID,
		  reqid.connSubID,
		  srcdude, ntohs(srcp),
		  dstdude, ntohs(dstp), fd);
}

void
connLogConnDone(conn_id_t reqid)
{
    if (LogfileStatus != LOG_ENABLE) return;
    if (reqid.connID == 0) return;
    ;
}

void
connLogSubConnDone(conn_id_t reqid)
{
    if (LogfileStatus != LOG_ENABLE) return;
    if (reqid.connID == 0) return;
    LOG_PREAMBLE();
    logfilePrintf(logfile,
		  "%ld.%6.6ld %lu subconndone %lu\n",
		  current_time.tv_sec, current_time.tv_usec, reqid.connID,
		  reqid.connSubID);
}

void
connLogConnPconn(conn_id_t reqid)
{
    if (LogfileStatus != LOG_ENABLE) return;
    if (reqid.connID == 0) return;
    LOG_PREAMBLE();
    logfilePrintf(logfile,
		  "%ld.%6.6ld %lu pconn\n",
		  current_time.tv_sec, current_time.tv_usec, reqid.connID);
}

void
connLogSubConnFailed(conn_id_t reqid, char *sys, char *msg)
{
    if (LogfileStatus != LOG_ENABLE) return;
    if (reqid.connID == 0) return;
    LOG_PREAMBLE();
    logfilePrintf(logfile,
		  "%ld.%6.6ld %lu %lu abort %s %s\n",
		  current_time.tv_sec, current_time.tv_usec, reqid.connID,
		  reqid.connSubID, sys, msg);
}

void
connLogSubConnFailedErrno(conn_id_t reqid, char *sys, int err)
{
    char *smesg;
    switch (err) {
    case ECONNREFUSED:   smesg = "refused"; break;
    case ECONNRESET:     smesg = "reset";   break;
    case ENETDOWN:
    case EHOSTUNREACH:
    case ENETUNREACH:    smesg = "unreach"; break;
    case EPROTONOSUPPORT: 
    case ENOBUFS:        smesg = "localfault"; break;
    case ETIMEDOUT:      smesg = "timeout"; break;
    default:	     smesg = "unknown";
    }

    connLogSubConnFailed(reqid, sys, smesg);
}

void
connLogConnFailed(conn_id_t reqid, char *sys, char *msg)
{
    if (LogfileStatus != LOG_ENABLE) return;
    if (reqid.connID == 0) return;
    LOG_PREAMBLE();
    logfilePrintf(logfile,
		  "%ld.%6.6ld %lu abort %s %s\n",
		  current_time.tv_sec, current_time.tv_usec, reqid.connID,
		  sys, msg);
}

void
connLogConnAbort(conn_id_t reqid, struct in_addr srca, u_short srcp,
		struct in_addr dsta, u_short dstp)
{
    if (LogfileStatus != LOG_ENABLE) return;
    if (reqid.connID == 0) return;
    ;
}

void
connLogRecvICP(conn_id_t reqid, struct in_addr srca, u_short srcp,
	       struct in_addr dsta, u_short dstp,
	       u_int32_t opcode, u_int32_t shostid, u_int32_t icp_reqnum,
	       u_int32_t flags, u_int32_t pad)
{
    char srcdude[64];
    char dstdude[64];

    if (LogfileStatus != LOG_ENABLE) return;
    /* We don't know the reqid yet, so it's allowed to be zero here */

    strcpy(srcdude, inet_ntoa(srca));
    strcpy(dstdude, inet_ntoa(dsta));

    LOG_PREAMBLE();
    logfilePrintf(logfile,
		  "%ld.%6.6ld %lu icprecv %lu %lu %s:%d %s:%d %x %x %x\n",
		  current_time.tv_sec, current_time.tv_usec, reqid.connID,
		  shostid, icp_reqnum,
		  srcdude, ntohs(srcp),
		  dstdude, ntohs(dstp),
		  opcode, flags, pad);
}

void
connLogGotICPQ(conn_id_t reqid, struct in_addr srca, u_short srcp,
	       struct in_addr dsta, u_short dstp, u_int32_t icp_src,
	       u_int32_t icp_reqnum)
{
    char srcdude[64];
    char dstdude[64];

    if (LogfileStatus != LOG_ENABLE) return;
    if (reqid.connID == 0) return;

    strcpy(srcdude, inet_ntoa(srca));
    strcpy(dstdude, inet_ntoa(dsta));

    LOG_PREAMBLE();
    logfilePrintf(logfile,
		  "%ld.%6.6ld %lu icpq %lu %lu %s:%d %s:%d\n",
		  current_time.tv_sec, current_time.tv_usec, reqid.connID,
		  icp_src, icp_reqnum,
		  srcdude, ntohs(srcp),
		  dstdude, ntohs(dstp));
}

void
connLogDupICPQ(conn_id_t reqid, conn_id_t rreqid)
{
    if (LogfileStatus != LOG_ENABLE) return;
    if (reqid.connID == 0) return;
    LOG_PREAMBLE();
    logfilePrintf(logfile,
		  "%ld.%6.6ld %lu icpdup %lu\n",
		  current_time.tv_sec, current_time.tv_usec, reqid.connID,
		  rreqid.connID);
}

void
connLogSendICP(conn_id_t reqid, struct in_addr srca, u_short srcp,
	       struct in_addr dsta, u_short dstp,
	       u_int32_t opcode, u_int32_t shostid, u_int32_t icp_reqnum,
	       u_int32_t flags, u_int32_t pad)
{
    char srcdude[64];
    char dstdude[64];

    if (LogfileStatus != LOG_ENABLE) return;
    if (reqid.connID == 0) return;

    strcpy(srcdude, inet_ntoa(srca));
    strcpy(dstdude, inet_ntoa(dsta));

    LOG_PREAMBLE();
    logfilePrintf(logfile,
		  "%ld.%6.6ld %lu icpsend %lu %lu %s:%d %s:%d %x %x %x\n",
		  current_time.tv_sec, current_time.tv_usec, reqid.connID,
		  shostid, icp_reqnum,
		  srcdude, ntohs(srcp),
		  dstdude, ntohs(dstp),
		  opcode, flags, pad);
}

void
connLogFetchViaPeer(conn_id_t reqid, struct in_addr srca, u_short srcp,
		    struct in_addr dsta, u_short dstp)
{
    char srcdude[64];
    char dstdude[64];

    if (LogfileStatus != LOG_ENABLE) return;
    if (reqid.connID == 0) return;

    strcpy(srcdude, inet_ntoa(srca));
    strcpy(dstdude, inet_ntoa(dsta));

    LOG_PREAMBLE();
    logfilePrintf(logfile,
		  "%ld.%6.6ld %lu fetchpeer %s:%d %s:%d\n",
		  current_time.tv_sec, current_time.tv_usec, reqid.connID,
		  srcdude, ntohs(srcp),
		  dstdude, ntohs(dstp));
}

void
connLogFetchViaSubconn(conn_id_t reqid)
{
    if (LogfileStatus != LOG_ENABLE) return;
    if (reqid.connID == 0) return;

    LOG_PREAMBLE();
    logfilePrintf(logfile,
		  "%ld.%6.6ld %lu fetchsubconn %lu\n",
		  current_time.tv_sec, current_time.tv_usec, reqid.connID,
		  reqid.connSubID);
}

void
connLogDNSCacheHit(conn_id_t reqid, char *status, char *name)
{
    if (LogfileStatus != LOG_ENABLE) return;
    if (reqid.connID == 0) return;
    LOG_PREAMBLE();
    logfilePrintf(logfile,
		  "%ld.%6.6ld %lu dns-cachehit %s %s\n",
		  current_time.tv_sec, current_time.tv_usec, reqid.connID,
		  status, name ? name : "(null)");
}

void
connLogSendDNS(conn_id_t reqid, int dns_reqno, struct in_addr dsta,
	       u_short dstp, char *name)
{
    if (LogfileStatus != LOG_ENABLE) return;
    if (reqid.connID == 0) return;
    LOG_PREAMBLE();
    logfilePrintf(logfile,
		  "%ld.%6.6ld %lu dns-send %d %s:%d %s\n",
		  current_time.tv_sec, current_time.tv_usec, reqid.connID,
		  dns_reqno, inet_ntoa(dsta), dstp,
		  name ? name : "(null name)");
}

void
connLogGotDNS(conn_id_t reqid, int dns_reqno, struct in_addr srca, u_short srcp)
{
    if (LogfileStatus != LOG_ENABLE) return;
    /* reqid can be zero because of late responses */
    LOG_PREAMBLE();
    logfilePrintf(logfile,
		  "%ld.%6.6ld %lu dns-recv %d %s:%d\n",
		  current_time.tv_sec, current_time.tv_usec, reqid.connID,
		  dns_reqno, inet_ntoa(srca), srcp);
    ;
}

void
connLogDNSFail(conn_id_t reqid, char *error)
{
    if (LogfileStatus != LOG_ENABLE) return;
    if (reqid.connID == 0) return;
    LOG_PREAMBLE();
    logfilePrintf(logfile,
		  "%ld.%6.6ld %lu dns-fail %s\n",
		  current_time.tv_sec, current_time.tv_usec, reqid.connID,
		  error);
}
