/*
 * $Id: forward.c,v 1.17 2004/08/09 18:48:36 dga Exp $
 *
 * DEBUG: section 17    Request Forwarding
 * AUTHOR: Duane Wessels
 *
 * SQUID Web Proxy Cache          http://www.squid-cache.org/
 * ----------------------------------------------------------
 *
 *  Squid is the result of efforts by numerous individuals from
 *  the Internet community; see the CONTRIBUTORS file for full
 *  details.   Many organizations have provided support for Squid's
 *  development; see the SPONSORS file for full details.  Squid is
 *  Copyrighted (C) 2001 by the Regents of the University of
 *  California; see the COPYRIGHT file for full details.  Squid
 *  incorporates software developed and/or copyrighted by other
 *  sources; see the CREDITS file for full details.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 *
 */


#include "squid.h"

static PSC fwdStartComplete;
static void fwdDispatch(FwdState *);
static void fwdConnectStart(void *);	/* should be same as EVH */
static void fwdStateFree(FwdState * fwdState);
static PF fwdConnectTimeout;
static PF fwdServerClosed;
static CNCB fwdConnectDone;
static int fwdCheckRetry(FwdState * fwdState);
static int fwdReforward(FwdState *);
static void fwdStartFail(FwdState *);
static void fwdLogReplyStatus(int tries, http_status status);
static OBJH fwdStats;
static STABH fwdAbort;
static peer *fwdStateServerPeer(FwdState *);

#define MAX_FWD_STATS_IDX 9
static int FwdReplyCodes[MAX_FWD_STATS_IDX + 1][HTTP_INVALID_HEADER + 1];

#if WIP_FWD_LOG
static void fwdLog(FwdState * fwdState);
static Logfile *logfile = NULL;
#endif

void pathSelect(FwdState *fwdState);
static void pathSelectDirectCheck(FwdState *fwdState);
static void getFailaddrsCallback(int mask, void *data);
static void pathSelectNeverDirectCheck(int mask, void *data);
static void pathSelectNeverDirectCheckDone(int mask, void *data);
static int pathSelectComputeConcurrency(failaddr *f, int n_dns);
void pathSelectDNSDone(const ipcache_addrs *ia, void *data);
void pathSelectSpecOpen(FwdState *fwdState);
void pathSelectSpecClosed(int fd, void *data);
void pathSelectSpecDone(int fd, int status, void *data);
void pathSelectSpecTimeout(int fd, void *data);
static void pathSelectICPTimeout(void *data);
static void pathSelectHoldTimeout(int fd, void *data);
static void pathSelectHoldConnect(int fd, int status, void *data);
static void pathSelectHoldClosed(int fd, void *data);
static void pathSelectHoldDNSDone(const ipcache_addrs *ia, void *data);
static void pathSelectStartICPConnects(FwdState *fwdState);
static void pathSelectHandlePingReply(peer * p, peer_t type, protocol_t proto, void *pingdata, void *data, struct sockaddr_in *to);
static void
pathSelectHandleICPReply(peer * p, peer_t type, icp_common_t * header,
			 void *data, struct sockaddr_in *to);
#if USE_HTCP
static void
pathSelectHandleHTCPReply(peer * p, peer_t type, htcpReplyData * header,
			 void *data, struct sockaddr_in *to);
#endif
void peerAddFwdServer(FwdServer ** FS, peer * p, hier_code code);
static void
fwdMultiConnectToPeer(FwdState *fwdState, const char *host, unsigned short port,
		      time_t timeout, CNCB *callback, FwdServer *fs);
static void fwdMultiConnectToPeerDNSDone(const ipcache_addrs *ia, void *data);
static void fwdMultiConnectToPeerTimeout(int fd, void *data);
static void fwdMultiConnectToPeerDone(int fd, int status, void *data);

static ipcache_addrs *
ipcache_addr_copy(const ipcache_addrs *ia)
{
    ipcache_addrs *n;
    int i;

    n = xcalloc(1, sizeof(ipcache_addrs));
    if (NULL == n) {
	debug(5, 1) ("ipcache_addr_copy:  No memory!\n");
	return NULL;
    }
    n->in_addrs = xcalloc(ia->count, sizeof(struct in_addr));
    n->bad_mask = xcalloc(ia->count, sizeof(unsigned char));
    
    if (NULL == n->in_addrs || NULL == n->bad_mask) {
	debug(5, 1) ("ipcache_addr_copy:  No memory for addrs!\n");
	if (n->in_addrs) safe_free(n->in_addrs);
	if (n->bad_mask) safe_free(n->bad_mask);
	safe_free(n);
	return NULL;
    }
    n->count = ia->count;
    n->cur = ia->cur;
    n->badcount = ia->badcount;
    for (i = 0; i < ia->count; i++) {
	n->in_addrs[i] = ia->in_addrs[i];
	n->bad_mask[i] = ia->bad_mask[i];
    }
    return n;
}

static void
ipcache_addr_copy_free(ipcache_addrs *ia)
{
    if (!ia) return;
    if (ia->in_addrs) safe_free(ia->in_addrs);
    if (ia->bad_mask) safe_free(ia->bad_mask);
    ia->in_addrs = NULL;
    safe_free(ia);
}

static peer *
fwdStateServerPeer(FwdState * fwdState)
{
    if (NULL == fwdState)
	return NULL;
    if (NULL == fwdState->servers)
	return NULL;
    return fwdState->servers->peer;
}

static void
fwdServerFree(FwdServer * fs)
{
    if (fs->peer)
	cbdataUnlock(fs->peer);
    memFree(fs, MEM_FWD_SERVER);
}

static void
fwdStateFree(FwdState * fwdState)
{
    StoreEntry *e = fwdState->entry;
    int sfd;
    int i;
    peer *p;
    debug(17, 3) ("fwdStateFree: %p\n", fwdState);
    assert(e->mem_obj);
#if URL_CHECKSUM_DEBUG
    assert(e->mem_obj->chksum == url_checksum(e->mem_obj->url));
#endif
#if WIP_FWD_LOG
    fwdLog(fwdState);
#endif
    if (e->mem_obj && e->mem_obj->ircb_data &&
	e->mem_obj->ircb_data == fwdState) {
	e->mem_obj->ircb_data = NULL;
	e->mem_obj->ping_reply_callback = NULL;
    }
    if (e->store_status == STORE_PENDING) {
	if (e->mem_obj->inmem_hi == 0) {
	    assert(fwdState->err);
	    errorAppendEntry(e, fwdState->err);
	    fwdState->err = NULL;
	} else {
	    EBIT_CLR(e->flags, ENTRY_FWD_HDR_WAIT);
	    storeComplete(e);
	    storeReleaseRequest(e);
	}
    }
    if (storePendingNClients(e) > 0)
	assert(!EBIT_TEST(e->flags, ENTRY_FWD_HDR_WAIT));
    p = fwdStateServerPeer(fwdState);
    fwdServersFree(&fwdState->servers);
    requestUnlink(fwdState->request);
    fwdState->request = NULL;
    if (fwdState->err)
	errorStateFree(fwdState->err);
    storeUnregisterAbort(e);
    storeUnlockObject(e);
    fwdState->entry = NULL;
    sfd = fwdState->server_fd;
    if (sfd > -1) {
	debug(17, 3) ("fwdStateFree:  removing close handler %d\n", sfd);
	comm_remove_close_handler(sfd, fwdServerClosed, fwdState);
	fwdState->server_fd = -1;
	debug(17, 3) ("fwdStateFree: closing FD %d\n", sfd);
	comm_close(sfd);
	if (p)
	    p->stats.conn_open--;
    }
    if (fwdState->reqhost_addr)
	ipcache_addr_copy_free(fwdState->reqhost_addr);
    for (i = 0; i < MAX_SPEC_OPENS; i++) {
	int fd;
	time_t to;
	if ((fd = fwdState->specOpens[i].fd) != -1) {
	    id_holder *h = cbdataAlloc(id_holder);
	    assert(h != NULL);
	    h->connID = fwdState->specOpens[i].connSubID;
	    cbdataLock(h); /* b/c commSetTimeout doesn't lock the handler */
	    to = fwdState->specOpens[i].timeout;
	    comm_remove_close_handler(fd, pathSelectSpecClosed, fwdState);
	    if (to > squid_curtime) {
		debug(17, 3) ("Resetting timeout fd %d to HoldTimeout in %lu\n", fd, to - squid_curtime);
		comm_add_close_handler(fd, pathSelectHoldClosed, h);
		commSetTimeout(fd, 
			       to - squid_curtime,
			       pathSelectHoldTimeout,
			       h);
		commSetCallback(fd, pathSelectHoldConnect, h);

	    }
	}
    }
    
    cbdataFree(fwdState);
}

static int
fwdCheckRetry(FwdState * fwdState)
{
    if (shutting_down)
	return 0;
    if (fwdState->entry->store_status != STORE_PENDING)
	return 0;
    if (fwdState->entry->mem_obj->inmem_hi > 0)
	return 0;
    if (fwdState->n_tries > 10)
	return 0;
    if (squid_curtime - fwdState->start > Config.Timeout.connect)
	return 0;
    if (fwdState->flags.dont_retry)
	return 0;
    if (fwdState->request->flags.body_sent)
	return 0;
    return 1;
}

static int
fwdCheckRetriable(FwdState * fwdState)
{
    /* If there is a request body then Squid can only try once
     * even if the method is indempotent
     */
    if (fwdState->request->body_connection)
	return 0;

    /* RFC2616 9.1 Safe and Idempotent Methods */
    switch (fwdState->request->method) {
	/* 9.1.1 Safe Methods */
    case METHOD_GET:
    case METHOD_HEAD:
	/* 9.1.2 Indepontent Methods */
    case METHOD_PUT:
    case METHOD_DELETE:
    case METHOD_OPTIONS:
    case METHOD_TRACE:
	break;
    default:
	return 0;
    }

    return 1;
}

static void
fwdServerClosed(int fd, void *data)
{
    FwdState *fwdState = data;
    debug(17, 2) ("fwdServerClosed: FD %d %s\n", fd, storeUrl(fwdState->entry));
    assert(fwdState->server_fd == fd);
    fwdState->server_fd = -1;
    if (fwdCheckRetry(fwdState)) {
	debug(17, 3) ("fwdServerClosed: re-forwarding (%d tries, %d secs)\n",
	    fwdState->n_tries,
	    (int) (squid_curtime - fwdState->start));
	if (fwdState->servers->next) {
	    /* use next, or cycle if origin server isn't last */
	    FwdServer *fs = fwdState->servers;
	    FwdServer **T, *T2 = NULL;
	    fwdState->servers = fs->next;
	    for (T = &fwdState->servers; *T; T2 = *T, T = &(*T)->next);
	    if (T2 && T2->peer) {
		/* cycle */
		*T = fs;
		fs->next = NULL;
	    } else {
		/* Use next. The last "direct" entry is retried multiple times */
		fwdState->servers = fs->next;
		fwdServerFree(fs);
	    }
	}
	/* use eventAdd to break potential call sequence loops */
	eventAdd("fwdConnectStart", fwdConnectStart, fwdState, 0.0, 0);
	return;
    }
    if (!fwdState->err && shutting_down) {
	fwdState->err = errorCon(ERR_SHUTTING_DOWN, HTTP_SERVICE_UNAVAILABLE);
	fwdState->err->request = requestLink(fwdState->request);
    }
    fwdStateFree(fwdState);
}

static void
fwdConnectDone(int server_fd, int status, void *data)
{
    FwdState *fwdState = data;
    static FwdState *current = NULL;
    FwdServer *fs = fwdState->servers;
    ErrorState *err;
    request_t *request = fwdState->request;
    assert(current != fwdState);
    current = fwdState;
    assert(fwdState->server_fd == server_fd);
    if (status == COMM_ERR_DNS) {
	/*
	 * Only set the dont_retry flag if the DNS lookup fails on
	 * a direct connection.  If DNS lookup fails when trying
	 * a neighbor cache, we may want to retry another option.
	 */
	if (NULL == fs->peer)
	    fwdState->flags.dont_retry = 1;
	debug(17, 4) ("fwdConnectDone: Unknown host: %s\n",
	    request->host);
	err = errorCon(ERR_DNS_FAIL, HTTP_SERVICE_UNAVAILABLE);
	err->dnsserver_msg = xstrdup(dns_error_message);
	err->request = requestLink(request);
	fwdState->connect_via = FWD_CONNECT_FAILED;
	fwdFail(fwdState, err);
	if (fs->peer)
	    fs->peer->stats.conn_open--;
	comm_close(server_fd);
    } else if (status != COMM_OK) {
	assert(fs);
	err = errorCon(ERR_CONNECT_FAIL, HTTP_SERVICE_UNAVAILABLE);
	err->xerrno = errno;
	if (fs->peer) {
	    err->host = xstrdup(fs->peer->host);
	    err->port = fs->peer->http_port;
	} else {
	    err->host = xstrdup(request->host);
	    err->port = request->port;
	}
	err->request = requestLink(request);
	fwdFail(fwdState, err);
	if (fs->peer) {
	    peerConnectFailed(fs->peer);
	    fs->peer->stats.conn_open--;
	}
	comm_close(server_fd);
    } else {
	debug(17, 3) ("fwdConnectDone: FD %d: '%s'\n", server_fd, storeUrl(fwdState->entry));
	if (fs->peer)
	    hierarchyNote(&fwdState->request->hier, fs->code, fs->peer->host);
	else if (Config.onoff.log_ip_on_direct)
	    hierarchyNote(&fwdState->request->hier, fs->code, fd_table[server_fd].ipaddr);
	else
	    hierarchyNote(&fwdState->request->hier, fs->code, request->host);
	fd_note(server_fd, storeUrl(fwdState->entry));
	fd_table[server_fd].uses++;
	if (fs->peer)
	    peerConnectSucceded(fs->peer);
	fwdDispatch(fwdState);
    }
    current = NULL;
}

static void
fwdConnectTimeout(int fd, void *data)
{
    FwdState *fwdState = data;
    StoreEntry *entry = fwdState->entry;
    ErrorState *err;
    peer *p = fwdStateServerPeer(fwdState);
    debug(17, 2) ("fwdConnectTimeout: FD %d: '%s'\n", fd, storeUrl(entry));
    assert(fd == fwdState->server_fd);
    if (entry->mem_obj->inmem_hi == 0) {
	err = errorCon(ERR_CONNECT_FAIL, HTTP_GATEWAY_TIMEOUT);
	err->request = requestLink(fwdState->request);
	err->xerrno = ETIMEDOUT;
	fwdFail(fwdState, err);
	/*
	 * This marks the peer DOWN ... 
	 */
	if (fwdState->servers)
	    if (fwdState->servers->peer)
		peerConnectFailed(fwdState->servers->peer);
    }
    if (p)
	p->stats.conn_open--;
    comm_close(fd);
}

static struct in_addr
aclMapAddr(acl_address * head, aclCheck_t * ch)
{
    acl_address *l;
    struct in_addr addr;
    for (l = head; l; l = l->next) {
	if (aclMatchAclList(l->acl_list, ch))
	    return l->addr;
    }
    addr.s_addr = INADDR_ANY;
    return addr;
}

static int
aclMapTOS(acl_tos * head, aclCheck_t * ch)
{
    acl_tos *l;
    for (l = head; l; l = l->next) {
	if (aclMatchAclList(l->acl_list, ch))
	    return l->tos;
    }
    return 0;
}

struct in_addr
getOutgoingAddr(request_t * request)
{
    aclCheck_t ch;
    memset(&ch, '\0', sizeof(aclCheck_t));
    if (request) {
	ch.src_addr = request->client_addr;
	ch.my_addr = request->my_addr;
	ch.my_port = request->my_port;
	ch.request = request;
    }
    return aclMapAddr(Config.accessList.outgoing_address, &ch);
}

unsigned long
getOutgoingTOS(request_t * request)
{
    aclCheck_t ch;
    memset(&ch, '\0', sizeof(aclCheck_t));
    if (request) {
	ch.src_addr = request->client_addr;
	ch.my_addr = request->my_addr;
	ch.my_port = request->my_port;
	ch.request = request;
    }
    return aclMapTOS(Config.accessList.outgoing_tos, &ch);
}

static void
fwdConnectStart(void *data)
{
    FwdState *fwdState = data;
    const char *url = storeUrl(fwdState->entry);
    int fd = -1;
    FwdServer *fs = fwdState->servers;
    const char *host;
    unsigned short port;
    time_t ctimeout;
    assert(fs);
    assert(fwdState->server_fd == -1);
    debug(17, 3) ("fwdConnectStart: %s\n", url);
    if (fs->peer) {
	host = fs->peer->host;
	port = fs->peer->http_port;
	ctimeout = fs->peer->connect_timeout > 0 ? fs->peer->connect_timeout
	    : Config.Timeout.peer_connect;
    } else if (fwdState->request->flags.accelerated &&
	Config.Accel.single_host && Config.Accel.host) {
	host = Config.Accel.host;
	port = Config.Accel.port;
	ctimeout = Config.Timeout.connect;
    } else {
	host = fwdState->request->host;
	port = fwdState->request->port;
	ctimeout = Config.Timeout.connect;
    }
    if (fwdCheckRetriable(fwdState)) {
	if ((fd = pconnPop(host, port)) >= 0) {
	    debug(17, 3) ("fwdConnectStart: reusing pconn FD %d\n", fd);
	    connLogConnPconn(fwdState->connID);
	    fwdState->server_fd = fd;
	    fwdState->n_tries++;
	    comm_add_close_handler(fd, fwdServerClosed, fwdState);
	    fwdConnectDone(fd, COMM_OK, fwdState);
	    return;
	}
    }
#if URL_CHECKSUM_DEBUG
    assert(fwdState->entry->mem_obj->chksum == url_checksum(url));
#endif
    /* We couldn't use a pconn.  We'll need to multiconnect directly
       to the peer (or server, in the event of something weird) */

    fwdMultiConnectToPeer(fwdState, host, port, ctimeout, fwdConnectDone, fs);
}

static void
fwdMultiConnectToPeer(FwdState *fwdState, const char *host, unsigned short port,
		      time_t timeout, CNCB *callback, FwdServer *fs)
{
    MCHolder *mcHolder;
    mcHolder = cbdataAlloc(MCHolder);
    mcHolder->status = MCHOLDER_START;
    mcHolder->fwdState = fwdState;
    mcHolder->host = host;
    mcHolder->port = port;
    mcHolder->timeout = timeout;
    mcHolder->callback = callback;
    mcHolder->fs = fs;
    cbdataLock(fwdState);
    
    ipcache_nbgethostbyname(host, fwdMultiConnectToPeerDNSDone,
			    mcHolder, connID_none);
}

static void
fwdMultiConnectToPeerDNSDone(const ipcache_addrs *ia, void *data)
{
    MCHolder *mcHolder = data;
    FwdState *fwdState = mcHolder->fwdState;
    ErrorState *err;
    unsigned short tos;
    failaddr *f = NULL;
    int i, concurrency;
    struct in_addr outgoing;
    int n_opened = 0;

    if (!cbdataValid(fwdState)) {
	cbdataUnlock(mcHolder->fwdState);
	cbdataFree(mcHolder);
	return;
    }

    if (NULL == ia) {
	debug(5, 3) ("fwdMultiConnectToPeerDNSDone:  Unknown host: %s\n",
		     mcHolder->host);
	if (!dns_error_message) {
	    dns_error_message = "Unknown DNS error";
	    debug(5, 1) ("commConnectDnsHandle: Bad dns_error_message\n");
	}
	assert(dns_error_message != NULL);
	cbdataUnlock(mcHolder->fwdState);
	mcHolder->callback(0, COMM_ERR_DNS, fwdState);
	cbdataFree(mcHolder);
	return;
    }

    mcHolder->reqhost_addr = ipcache_addr_copy(ia);
    tos = getOutgoingTOS(fwdState->request);
    concurrency = pathSelectComputeConcurrency(fwdState->failaddrs,
					       mcHolder->reqhost_addr->count);
    if (concurrency > MAX_SPEC_OPENS) {
	concurrency = MAX_SPEC_OPENS;
    }
    mcHolder->concurrency = concurrency;
    cbdataLock(mcHolder);
    for (i = 0; i < concurrency; i++) {
	int fd;
	if (!cbdataValid(mcHolder) || mcHolder->status != MCHOLDER_START) {
	    break;
	}
	if (f == NULL) f = fwdState->failaddrs;
	if (f != NULL) {
	    outgoing = f->addr;
	    f = f->next;
	}
	fd = comm_openex(SOCK_STREAM,
			 0,
			 outgoing,
			 0,
			 COMM_NONBLOCKING,
			 tos,
			 storeUrl(fwdState->entry));
	if (fd < 0) {
	    debug(50, 4) ("fwdMultiConnectToPeerDNSDone: %s\n", xstrerror());
	} else {
	    fwdState->spec_fds[mcHolder->n_open] = fd;
	    cbdataLock(mcHolder);
	    commSetTimeout(fd, Config.Timeout.connect,
			   fwdMultiConnectToPeerTimeout, mcHolder);
	    n_opened++;
	    mcHolder->n_open++;
	    commConnectStartExt(fd, mcHolder->host, mcHolder->port,
			    mcHolder->reqhost_addr,
			    fwdMultiConnectToPeerDone,
			    mcHolder,
			    connID_none);
	}
    }
    if (n_opened <= 0) {
	/* We couldn't allocate any file descriptors. */
	err = errorCon(ERR_SOCKET_FAILURE, HTTP_INTERNAL_SERVER_ERROR);
	err->xerrno = errno;
	err->request = requestLink(fwdState->request);
	fwdFail(fwdState, err);
	cbdataUnlock(mcHolder->fwdState);
	fwdStateFree(fwdState);
	cbdataFree(mcHolder);
    }
    cbdataUnlock(mcHolder);
}

static void
fwdMultiConnectToPeerCleanup(MCHolder *mcHolder)
{
    if (mcHolder->reqhost_addr)
	ipcache_addr_copy_free(mcHolder->reqhost_addr);

    cbdataUnlock(mcHolder->fwdState);
    cbdataFree(mcHolder);
}

static void
fwdMultiConnectToPeerTimeout(int fd, void *data)
{
    fwdMultiConnectToPeerDone(fd, COMM_TIMEOUT, data);
}

static void
fwdMultiConnectToPeerDone(int fd, int status, void *data)
{
    MCHolder *mcHolder = data;
    FwdState *fwdState = mcHolder->fwdState;
    FwdServer *fs;
    int i;

    cbdataUnlock(mcHolder); /* Counteract timeout */
    assert(mcHolder->n_open >= 1);
    mcHolder->n_open--;
    for (i = 0; i < mcHolder->concurrency; i++) {
	if (fwdState->spec_fds[i] == fd)
	    fwdState->spec_fds[i] = -1;
    }
    if (mcHolder->status != MCHOLDER_START) {
	if (0 == mcHolder->n_open) fwdMultiConnectToPeerCleanup(mcHolder);
	comm_close(fd);
	return;
    }
    if (status == COMM_OK) {
	mcHolder->status = MCHOLDER_DONE;
	fwdState->server_fd = fd;
	fwdState->n_tries++;
	/*
	 * stats.conn_open is used to account for the number of
	 * connections that we have open to the peer, so we can limit
	 * based on the max-conn option.  We need to increment here,
	 * even if the connection may fail.
	 */
	fs = mcHolder->fs;
	if (fs && fs->peer)
	    fs->peer->stats.conn_open++;
	comm_add_close_handler(fd, fwdServerClosed, fwdState);
	commSetTimeout(fd,
		       mcHolder->timeout,
		       fwdConnectTimeout,
		       fwdState);
	fwdConnectDone(fd, status, fwdState);
    } else {
	if (0 == mcHolder->n_open) {
	    mcHolder->status = MCHOLDER_FAILED;
	    fwdState->server_fd = fd;
	    fwdState->n_tries++;
	    comm_add_close_handler(fd, fwdServerClosed, fwdState);
	    fwdConnectDone(fd, status, fwdState);
	} else {
	    comm_close(fd);
	    /* ignore the first few errors */
	}
    }

    if (0 == mcHolder->n_open)
	fwdMultiConnectToPeerCleanup(mcHolder);
}

static void
fwdStartComplete(FwdServer * servers, void *data)
{
    FwdState *fwdState = data;
    debug(17, 3) ("fwdStartComplete: %s\n", storeUrl(fwdState->entry));
    if (servers != NULL) {
	fwdState->servers = servers;
	fwdConnectStart(fwdState);
    } else {
	fwdStartFail(fwdState);
    }
}

static void
fwdStartFail(FwdState * fwdState)
{
    ErrorState *err;
    debug(17, 3) ("fwdStartFail: %s\n", storeUrl(fwdState->entry));
    err = errorCon(ERR_CANNOT_FORWARD, HTTP_SERVICE_UNAVAILABLE);
    err->request = requestLink(fwdState->request);
    err->xerrno = errno;
    fwdFail(fwdState, err);
    fwdStateFree(fwdState);
}

static void
fwdDispatch(FwdState * fwdState)
{
    peer *p = NULL;
    request_t *request = fwdState->request;
    StoreEntry *entry = fwdState->entry;
    ErrorState *err;
    debug(17, 3) ("fwdDispatch: FD %d: Fetching '%s %s'\n",
	fwdState->client_fd,
	RequestMethodStr[request->method],
	storeUrl(entry));
    /*assert(!EBIT_TEST(entry->flags, ENTRY_DISPATCHED)); */
    assert(entry->ping_status != PING_WAITING);
    assert(entry->lock_count);
    EBIT_SET(entry->flags, ENTRY_DISPATCHED);
    netdbPingSite(request->host);
    /*
     * Assert that server_fd is set.  This is to guarantee that fwdState
     * is attached to something and will be deallocated when server_fd
     * is closed.
     */
    assert(fwdState->server_fd > -1);
    if (fwdState->servers && (p = fwdState->servers->peer)) {
	p->stats.fetches++;
	fwdState->request->peer_login = p->login;
	httpStart(fwdState);
    } else {
	fwdState->request->peer_login = NULL;
	switch (request->protocol) {
	case PROTO_HTTP:
	    httpStart(fwdState);
	    break;
	case PROTO_GOPHER:
	    gopherStart(fwdState);
	    break;
	case PROTO_FTP:
	    ftpStart(fwdState);
	    break;
	case PROTO_WAIS:
	    waisStart(fwdState);
	    break;
	case PROTO_CACHEOBJ:
	case PROTO_INTERNAL:
	case PROTO_URN:
	    fatal_dump("Should never get here");
	    break;
	case PROTO_WHOIS:
	    whoisStart(fwdState);
	    break;
	default:
	    debug(17, 1) ("fwdDispatch: Cannot retrieve '%s'\n",
		storeUrl(entry));
	    err = errorCon(ERR_UNSUP_REQ, HTTP_BAD_REQUEST);
	    err->request = requestLink(request);
	    fwdFail(fwdState, err);
	    /*
	     * Force a persistent connection to be closed because
	     * some Netscape browsers have a bug that sends CONNECT
	     * requests as GET's over persistent connections.
	     */
	    request->flags.proxy_keepalive = 0;
	    /*
	     * Set the dont_retry flag becuase this is not a
	     * transient (network) error; its a bug.
	     */
	    fwdState->flags.dont_retry = 1;
	    /*
	     * this assertion exists because if we are connected to
	     * a peer, then we need to decrement p->stats.conn_open.
	     */
	    assert(NULL == p);
	    comm_close(fwdState->server_fd);
	    break;
	}
    }
}

static int
fwdReforward(FwdState * fwdState)
{
    StoreEntry *e = fwdState->entry;
    FwdServer *fs = fwdState->servers;
    http_status s;
    assert(e->store_status == STORE_PENDING);
    assert(e->mem_obj);
#if URL_CHECKSUM_DEBUG
    assert(e->mem_obj->chksum == url_checksum(e->mem_obj->url));
#endif
    debug(17, 3) ("fwdReforward: %s?\n", storeUrl(e));
    if (!EBIT_TEST(e->flags, ENTRY_FWD_HDR_WAIT)) {
	debug(17, 3) ("fwdReforward: No, ENTRY_FWD_HDR_WAIT isn't set\n");
	return 0;
    }
    if (fwdState->n_tries > 9)
	return 0;
    if (fwdState->request->flags.body_sent)
	return 0;
    assert(fs);
    fwdState->servers = fs->next;
    fwdServerFree(fs);
    if (fwdState->servers == NULL) {
	debug(17, 3) ("fwdReforward: No forward-servers left\n");
	return 0;
    }
    s = e->mem_obj->reply->sline.status;
    debug(17, 3) ("fwdReforward: status %d\n", (int) s);
    return fwdReforwardableStatus(s);
}

/* PUBLIC FUNCTIONS */

void
fwdServersFree(FwdServer ** FS)
{
    FwdServer *fs;
    while ((fs = *FS)) {
	*FS = fs->next;
	fwdServerFree(fs);
    }
}

void
fwdStart(int fd, StoreEntry * e, request_t * r)
{
    FwdState *fwdState;
    aclCheck_t ch;
    int answer;
    ErrorState *err;
    int i;
    /*
     * client_addr == no_addr indicates this is an "internal" request
     * from peer_digest.c, asn.c, netdb.c, etc and should always
     * be allowed.  yuck, I know.
     */
    if (r->client_addr.s_addr != no_addr.s_addr) {
	/*      
	 * Check if this host is allowed to fetch MISSES from us (miss_access)
	 */
	memset(&ch, '\0', sizeof(aclCheck_t));
	ch.src_addr = r->client_addr;
	ch.my_addr = r->my_addr;
	ch.my_port = r->my_port;
	ch.request = r;
	answer = aclCheckFast(Config.accessList.miss, &ch);
	if (answer == 0) {
	    err = errorCon(ERR_FORWARDING_DENIED, HTTP_FORBIDDEN);
	    err->request = requestLink(r);
	    err->src_addr = r->client_addr;
	    errorAppendEntry(e, err);
	    return;
	}
    }
    debug(17, 3) ("fwdStart: '%s'\n", storeUrl(e));
    e->mem_obj->request = requestLink(r);
#if URL_CHECKSUM_DEBUG
    assert(e->mem_obj->chksum == url_checksum(e->mem_obj->url));
#endif
    if (shutting_down) {
	/* more yuck */
	err = errorCon(ERR_SHUTTING_DOWN, HTTP_SERVICE_UNAVAILABLE);
	err->request = requestLink(r);
	errorAppendEntry(e, err);
	return;
    }
    switch (r->protocol) {
	/*
	 * Note, don't create fwdState for these requests
	 */
    case PROTO_INTERNAL:
	internalStart(r, e);
	return;
    case PROTO_CACHEOBJ:
	cachemgrStart(fd, r, e);
	return;
    case PROTO_URN:
	urnStart(r, e);
	return;
    default:
	break;
    }
    fwdState = cbdataAlloc(FwdState);
    fwdState->entry = e;
    fwdState->client_fd = fd;
    fwdState->server_fd = -1;
    for (i = 0; i < MAX_SPEC_OPENS; i++)
	fwdState->specOpens[i].fd = -1;
    fwdState->request = requestLink(r);
    fwdState->start = squid_curtime;
    fwdState->connID = connLogNewClientReq(storeUrl(e),
					   r->flags.onecache,
					   r->client_addr);
    storeLockObject(e);
    EBIT_SET(e->flags, ENTRY_FWD_HDR_WAIT);
    storeRegisterAbort(e, fwdAbort, fwdState);
    cbdataLock(fwdState);
    /* The locking is a temporary hack b/c we don't control what
     * happens in path select, and it's possible for the fwdState
     * to get invalidated before peerselect locks it.  Not sure
     * this is actually a sufficient response, but.. wtf. */
#if 1
    pathSelect(fwdState);
#else
    if (cbdataValid(fwdState)) {
	peerSelect(r, e, fwdStartComplete, fwdState);
	fprintf(stderr, "peerSelect is done\n");
    }
#endif
    cbdataUnlock(fwdState);
}

/*
 * pathSelect - perform ACL matching to determine to whom
 * we can possibly forward our requests.  We let this
 * step complete before we actually initiate any packets.
 * If you want this step to be fast, don't do any ACL tests
 * that require DNS lookups...
 */

void
pathSelect(FwdState *fwdState)
{
    if (Config.accessList.failaddrs) {
	fwdState->cur_aclfail = Config.accessList.failaddrs;
	fwdState->acl_checklist = aclChecklistCreate(
	    fwdState->cur_aclfail->acl_list,
	    fwdState->request, NULL);
	aclNBCheck(fwdState->acl_checklist, getFailaddrsCallback, fwdState);
    } else {
	fwdState->failaddrs = NULL;
	pathSelectDirectCheck(fwdState);
    }
}

static void
getFailaddrsCallback(int mask, void *data)
{
    FwdState *fwdState = data;
    
    if (mask == 1) {
	failaddr *f;
	fwdState->acl_checklist = NULL;
	fwdState->failaddrs = fwdState->cur_aclfail->f;
	fwdState->failaddr_ct = 0;
	f = fwdState->failaddrs;
	while (f != NULL) {
	    f = f->next;
	    fwdState->failaddr_ct++;
	}
	pathSelectDirectCheck(fwdState);
    }
    else {
	if (NULL == fwdState->cur_aclfail->next) {
	    fwdState->failaddrs = NULL;
	    pathSelectDirectCheck(fwdState);
	} else {
	    fwdState->cur_aclfail = fwdState->cur_aclfail->next;
	    fwdState->acl_checklist = aclChecklistCreate(
	       fwdState->cur_aclfail->acl_list,
	       fwdState->request,
	       NULL);
	    aclNBCheck(fwdState->acl_checklist, getFailaddrsCallback,
		       fwdState);
	}
    }
}


static void
pathSelectDirectCheck(FwdState *fwdState)
{
    if (Config.accessList.AlwaysDirect) {
	fwdState->acl_checklist = aclChecklistCreate(
	    Config.accessList.AlwaysDirect,
	    fwdState->request,
	    NULL);
	aclNBCheck(fwdState->acl_checklist, pathSelectNeverDirectCheck,
		   fwdState);
    } else {
	pathSelectNeverDirectCheck(0, fwdState);
    }
}


static void
pathSelectNeverDirectCheck(int mask, void *data)
{
    FwdState *fwdState = data;
    if (mask) {
	fwdState->use_peers = FWD_NEVER_PEERS;
    } else {
	fwdState->use_peers = FWD_MAYBE_PEERS;
    }
    if (mask || !Config.accessList.NeverDirect) {
	fwdState->acl_checklist = NULL;
	pathSelectNeverDirectCheckDone(0, fwdState);
    } else {
	fwdState->acl_checklist = aclChecklistCreate(
	   Config.accessList.NeverDirect,
	   fwdState->request,
	   NULL);
	aclNBCheck(fwdState->acl_checklist,
		   pathSelectNeverDirectCheckDone,
		   fwdState);
    }
}

static void
pathSelectNeverDirectCheckDone(int mask, void *data)
{
    FwdState *fwdState = data;
#if 0
    char *fnp;
#endif

    if (mask) {
	fwdState->use_peers = FWD_ALWAYS_PEERS;
    }

#if 0
    switch(fwdState->use_peers) {
    case FWD_MAYBE_PEERS:  fnp = "maybe "; break;
    case FWD_NEVER_PEERS:  fnp = "never "; break;
    case FWD_ALWAYS_PEERS:  fnp = "always"; break;
    default: fnp = "ERROR "; break;
    }

    fprintf(stderr, "psndcd: use peers %s %s\n", fnp,
	    storeUrl(fwdState->entry));
#endif

    /* XXX:  Eventually, support for things like cache digests
     * should go here.  If we hit in the cache digest, then
     * treat as a proxy hit... maybe. :)  But worry about it
     * later. */

    /* The logic behind this ordering:
     * a)  If we get a peer cache hit, it will return quite fast.
     *     great!  use it.  If it takes forever for us to learn of
     *     the peer hit, perhaps it's not worth using.  There
     *     may bec
     */

    cbdataLock(fwdState);

    if (fwdState->use_peers != FWD_ALWAYS_PEERS &&
	fwdCheckRetriable(fwdState)) {
	int pfd;
	
	if ((pfd = pconnPop(fwdState->request->host, fwdState->request->port)) >= 0) {
	    debug(17, 3) ("pathSelectNeverDirectCheckDone:  reusing pconn FD %d\n", pfd);
	    connLogConnPconn(fwdState->connID);
	    /* This might make us overly dependent on the pconn... */
	    peerAddFwdServer(&fwdState->servers, NULL, DIRECT);
	    fwdState->server_fd = pfd;
	    fwdState->n_tries++;
	    comm_add_close_handler(pfd, fwdServerClosed, fwdState);
	    fwdConnectDone(pfd, COMM_OK, fwdState);
	    cbdataUnlock(fwdState);
	    return;
	}
    }
    
    if (fwdState->use_peers != FWD_NEVER_PEERS) {
	cbdataLock(fwdState);
	pathSelectStartICPConnects(fwdState);
    }

    if (fwdState->use_peers != FWD_ALWAYS_PEERS) {
	/* DNSDone leads into opening (perhaps multiple) connections
	 * to the origin server */
#if 1
	id_holder *h = cbdataAlloc(id_holder);
	cbdataLock(fwdState); /* unlock in pshdd */
#endif
	getCurrentTime();
	fwdState->request->perf.dns_start_d = current_dtime;
#if 1
	assert(h != NULL);
	h->connID = fwdState->connID;
	h->fwdState = fwdState;
	h->failaddrs = fwdState->failaddrs;
	xmemcpy(h->host, fwdState->request->host, SQUIDHOSTNAMELEN+1);
	h->port = fwdState->request->port;
	ipcache_nbgethostbyname(fwdState->request->host, pathSelectHoldDNSDone,
				h, fwdState->connID);
#else
	ipcache_nbgethostbyname(fwdState->request->host, pathSelectDNSDone,
	   fwdState, fwdState->connID);
#endif
    }
    cbdataUnlock(fwdState);
}

void
pathSelectDNSDone(const ipcache_addrs *ia, void *data)
{
    FwdState *fwdState = data;
    if (!cbdataValid(fwdState)) return;
    if (fwdState->connect_via != FWD_CONNECT_UNKNOWN) return;
    
    /* What should we do to the entire connection if the DNS
     * fails before we hear from our peers?  I think we
     * should bail, since it's an actively "no go" response.
     */
    getCurrentTime();
    debug(17, 3) ("pathSelectDNSDone: %s %s\n",
		  fwdState->request->host,
		  ia == NULL ? "failed" : "done");

    if (!ia) {
	if (dns_error_message &&
	    !strncmp(dns_error_message, "Server Failure", 14) &&
	    fwdState->icp_retries > 0 &&
	    fwdState->icp_retries < MAX_ICP_RETRIES) {
	    
	    /* This may just mean a non-answer, or worse, negative caching
	     * of a non-answer.  Let's wait a little bit for possible
	     * ICP replies before naking everything */
	    eventAdd("DelayedICPCB", pathSelectICPTimeout, fwdState,
		     (double)1, 1);
	} else {
	    ErrorState *err;
	    err = errorCon(ERR_DNS_FAIL, HTTP_SERVICE_UNAVAILABLE);
	    err->dnsserver_msg = xstrdup(dns_error_message);
	    err->request = requestLink(fwdState->request);
	    fwdState->connect_via = FWD_CONNECT_FAILED;
	    connLogConnFailed(fwdState->connID, "dns", err->dnsserver_msg);
	    fwdFail(fwdState, err);
	    fwdStateFree(fwdState);
	}
	return;
    }

    /* Log time */

    assert(ia->cur < ia->count);
    fwdState->reqhost_addr = ipcache_addr_copy(ia);
    fwdState->request->perf.dns_done_d = current_dtime;

    /* Otherwise, if not pconnable */
    pathSelectSpecOpen(fwdState);
}

static int
pathSelectComputeConcurrency(failaddr *f, int n_dns)
{
    int n_failaddrs = 0;
    
    while (f) {
	n_failaddrs++;
	f = f->next;
    }
    if (!n_failaddrs) n_failaddrs = 1;
    if (!n_dns) {
	n_dns = 1;
    }
    switch(n_failaddrs * n_dns) {
    case 1:
	return 1;
	break;
    case 2:
	return 2;
	break;
    default:
	return MAX_SPEC_OPENS;
    }
}

void
pathSelectSpecOpen(FwdState *fwdState)
{
    struct in_addr outgoing;
    unsigned short tos;
    int fd;
    int i;
    failaddr *f = NULL;
    
    if (fwdState->connect_via != FWD_CONNECT_UNKNOWN) return;
    /* This function should add as many speculative
     * connections as we want it to, and manage the
     * timer for adding more.  We'll return again and again
     * to this function to do magic. */

    if (!fwdState->concurrency) 
	fwdState->concurrency = pathSelectComputeConcurrency(fwdState->failaddrs,
							     fwdState->reqhost_addr->count);

    
    /* Check to see if we're out of speculative connection
     * slots. */
    if (fwdState->open_specs >= MAX_SPEC_OPENS) {
	return;
    }
    if (fwdState->open_specs >= fwdState->concurrency) {
	return;
    }
    if (fwdState->n_spec_tries >= MAX_SPEC_TRIES) {
	/* XXX:  Dump out an error */
	return;
    }
    fwdState->n_spec_tries++;
    /* Find a slot */
    for (i = 0; i < MAX_SPEC_OPENS; i++) {
	if (fwdState->specOpens[i].fd == -1) {
	    break;
	}
    }

    assert (i < MAX_SPEC_OPENS);
    /* This lock will force the TCP connection to stay open until the
     * connection times out... this is evil, but we need it for
     * logging properly */
    outgoing = getOutgoingAddr(fwdState->request);
    tos = getOutgoingTOS(fwdState->request);

    /* this is a super cheesy non combinatorial way of walking
     * through the addresses, but it's a start... */
    if (fwdState->failaddrs) {
	int ct = fwdState->spec_iter;
	f = fwdState->failaddrs;
	while (ct-- > 0) {
	    f = f->next;
	    if (f == NULL) f = fwdState->failaddrs;
	}
	outgoing = f->addr;
    }
    
    getCurrentTime();
    
    if (fwdState->request->perf.conn_start_d == 0)
	fwdState->request->perf.conn_start_d = current_dtime;
    
    fd = comm_openex(SOCK_STREAM, 0, outgoing, 0, COMM_NONBLOCKING,
		     tos, storeUrl(fwdState->entry));
    if (fd < 0) {
	ErrorState *err;
	debug(17, 3) ("pathSelectSpecOpen: %s\n", xstrerror());
	/* Is this a condition on which we should fail? */
	err = errorCon(ERR_SOCKET_FAILURE, HTTP_INTERNAL_SERVER_ERROR);
	err->xerrno = errno;
	err->request = requestLink(fwdState->request);
	fwdFail(fwdState, err);
	fwdStateFree(fwdState);
	return;
    }
    debug(17, 3) ("pathSelectSpecOpen %d %s\n", fd, storeUrl(fwdState->entry));
    fwdState->specOpens[i].fd = fd;
    fwdState->open_specs++;
    fwdState->spec_iter++;
    cbdataLock(fwdState); /* Setting the timeout doesn't lock the fwdState */
    comm_add_close_handler(fd, pathSelectSpecClosed, fwdState);
    fwdState->specOpens[i].timeout = 
	commSetTimeout(fd, Config.Timeout.connect, pathSelectSpecTimeout,
		       fwdState);
    debug(17, 3) ("Now %lu timeout at %lu\n", squid_curtime,
		  fwdState->specOpens[i].timeout);
    fwdState->specOpens[i].connSubID = fwdState->connID;
    connLogNewSub(&fwdState->specOpens[i].connSubID);
    commConnectStartExt(fd,
			fwdState->request->host,
			fwdState->request->port,
			fwdState->reqhost_addr,
			pathSelectSpecDone,
			fwdState,
			fwdState->specOpens[i].connSubID);
    /* for now, just open one per outbound connection */
    if (f && fwdState->open_specs < fwdState->failaddr_ct) {
	if (f->timeout > 2) {
	    eventAdd("DelayedFailCB", pathSelectSpecOpen, fwdState,
		     (double)(f->timeout) / 1000, 1);
	} else {
	    pathSelectSpecOpen(fwdState);
	}
    }
}

static void
pathSelectICPTimeout(void *data)
{
    FwdState *fwdState = data;
    StoreEntry *entry = fwdState->entry;

    if (fwdState->connect_via != FWD_CONNECT_UNKNOWN) return;

    if (cbdataValid(fwdState) && entry->mem_obj->inmem_hi == 0) {
	ErrorState *err;
	err = errorCon(ERR_CONNECT_FAIL, HTTP_SERVICE_UNAVAILABLE);
	err->request = requestLink(fwdState->request);
	err->xerrno = ETIMEDOUT;
	fwdState->connect_via = FWD_CONNECT_FAILED;
	connLogConnFailed(fwdState->connID, "dns", err->dnsserver_msg);
	fwdFail(fwdState, err);
	fwdStateFree(fwdState);
    }
}

void
pathSelectSpecTimeout(int fd, void *data)
{
    FwdState *fwdState = data;
    StoreEntry *entry = fwdState->entry;
    ErrorState *err;
    int spec_fd_no;
    conn_id_t subConnID = connID_none;

    debug(17, 3) ("pathSelectSpecTimeout on %d\n", fd);

    for (spec_fd_no = 0; spec_fd_no < MAX_SPEC_OPENS; spec_fd_no++) {
	if (fwdState->specOpens[spec_fd_no].fd == fd) {
	    subConnID = fwdState->specOpens[spec_fd_no].connSubID;
	    break;
	}
    }
    
    cbdataLock(fwdState);
    /* comm_close will release a lock on it.  We need to be sure
     * it stays around */
    if (fd != -1) comm_close(fd);
    
    if (!subConnID.connID) {
	debug(17, 1) ("pathSelectSpecTimeout no subconn fd %d (%lu) valid: %d convia: %d\n",
		      fd, fwdState->connID.connID, cbdataValid(fwdState),
		      cbdataValid(fwdState) ? fwdState->connect_via : 99999);
    }
    connLogSubConnFailed(subConnID, "tcp", "timeout");

    if (cbdataValid(fwdState)) {
	if (fwdState->n_spec_tries >= MAX_SPEC_TRIES && 
	    entry->mem_obj->inmem_hi == 0) {
	    err = errorCon(ERR_CONNECT_FAIL, HTTP_GATEWAY_TIMEOUT);
	    err->request = requestLink(fwdState->request);
	    err->xerrno = ETIMEDOUT;
	    fwdFail(fwdState, err);
	    fwdStateFree(fwdState);
	} else {
	    if (fd != -1) pathSelectSpecOpen(fwdState);
	}
    }
    cbdataUnlock(fwdState);
}

void
pathSelectSpecDone(int fd, int status, void *data)
{
    FwdState *fwdState = data;
    int spec_fd_no;
    conn_id_t subConnID;

    getCurrentTime();
    for (spec_fd_no = 0; spec_fd_no < MAX_SPEC_OPENS; spec_fd_no++) {
	if (fwdState->specOpens[spec_fd_no].fd == fd) {
	    subConnID = fwdState->specOpens[spec_fd_no].connSubID;
	    break;
	}
    }

    if (fwdState->connect_via != FWD_CONNECT_UNKNOWN) {
	/* Someone's already succeded.  Close. */
	comm_close(fd);
	return;
    }

    assert (status != COMM_ERR_DNS); /* already trapped */
    if (status != COMM_OK) {
	debug(17, 3) ("Failed speculative connection on fd %d\n", fd);
	comm_close(fd);
	if (cbdataValid(fwdState)) {
	    StoreEntry *entry = fwdState->entry;
	    ErrorState *err;
	    if (fwdState->n_spec_tries >= MAX_SPEC_TRIES && 
		entry->mem_obj->inmem_hi == 0) {
		err = errorCon(ERR_CONNECT_FAIL, HTTP_GATEWAY_TIMEOUT);
		err->request = requestLink(fwdState->request);
		err->xerrno = ETIMEDOUT;
		fwdFail(fwdState, err);
		fwdStateFree(fwdState);
	    } else {
		if (fd != -1) pathSelectSpecOpen(fwdState);
	    }
	}
	return;
    }
    /* XXX:  Just changed to force the error condition to call
     * the close handler for cleanup..  is this proper? */
    if (spec_fd_no < MAX_SPEC_OPENS) {
	fwdState->specOpens[spec_fd_no].fd = -1;
    }

    
    debug(17, 3) ("Successful speculative connection on fd %d\n", fd);
    fwdState->request->perf.conn_done_d = current_dtime;
    fwdState->connect_via = FWD_CONNECT_LOCAL;
    peerAddFwdServer(&fwdState->servers, NULL, DIRECT);
    fwdState->server_fd = fd;
    fwdState->n_tries++;
    
    comm_remove_close_handler(fd, pathSelectSpecClosed, fwdState);
    comm_add_close_handler(fd, fwdServerClosed, fwdState);
    
    fwdState->entry->ping_status = PING_DONE;
    connLogFetchViaSubconn(subConnID);
    fwdConnectDone(fd, COMM_OK, fwdState);
}

void
pathSelectHoldOpen(id_holder *h)
{
    struct in_addr outgoing;
    unsigned short tos;
    int fd;
    failaddr *f = NULL;
    int concurrency;
    id_holder *sh;
    int n_open = 0;

    concurrency = pathSelectComputeConcurrency(h->failaddrs,
					       h->reqhost_addr->count);

    debug(17, 3) ("pathSelectHoldOpen starting\n");
    bzero(&outgoing, sizeof(outgoing));
    outgoing = getOutgoingAddr(NULL);
    tos = getOutgoingTOS(NULL);

    if (h->failaddrs) f = h->failaddrs;

    while (n_open < concurrency) {
	if (h->failaddrs) {
	    outgoing = f->addr;
	    f = f->next;
	    if (f == NULL) f = h->failaddrs;
	}
    
	getCurrentTime();
    
	fd = comm_openex(SOCK_STREAM, 0, outgoing, 0, COMM_NONBLOCKING,
			 tos, "deferred test");

	if (fd < 0) {
	    debug(17, 1) ("fuck:  openex on a HOLD connection failed\n");
	} else {
	    /* XXX:  Create a _new_ h to hold the thing */
	    sh = cbdataAlloc(id_holder);
	    assert (sh != NULL);
	    cbdataLock(sh);
	    sh->connID = h->connID;
	    connLogNewSub(&sh->connID);
	    comm_add_close_handler(fd, pathSelectHoldClosed, sh);
	    commSetTimeout(fd, Config.Timeout.connect, pathSelectHoldTimeout, sh);
	    commConnectStartExt(fd,
				h->host,
				h->port,
				h->reqhost_addr,
				pathSelectHoldConnect,
				sh,
				sh->connID);
	}
	n_open++;
    }
}


static void
pathSelectHoldDNSDone(const ipcache_addrs *ia, void *data)
{
    /* If we still have a valid object, just dispatch to it */
    id_holder *h = data;
    FwdState *f = h->fwdState;

    if (cbdataValid(f) &&
        f->connect_via == FWD_CONNECT_UNKNOWN) {
	cbdataFree(h);
	pathSelectDNSDone(ia, f);
	cbdataUnlock(f);
	return;
    }

    if (!ia) {
	connLogDNSFail(h->connID, dns_error_message);
	cbdataUnlock(f);
	cbdataFree(h);
	return;
    }

    /* XXX - no error, but our fwdState is already ocupied.
     * Toss off some speculative connections and set them to
     * just return to the holder procs. */
    assert(ia->cur < ia->count);
    h->reqhost_addr = ipcache_addr_copy(ia);

    /* Otherwise, if not pconnable */
    pathSelectHoldOpen(h);
    cbdataUnlock(f);
    ipcache_addr_copy_free(h->reqhost_addr);
    cbdataFree(h);
}

static void
pathSelectHoldClosed(int fd, void *data)
{
    debug(17, 1) ("pathSelectHoldClosed unexpected\n");
}

static void
pathSelectHoldTimeout(int fd, void *data)
{
    id_holder *h = data;
    
    getCurrentTime();
    
    comm_remove_close_handler(fd, pathSelectHoldClosed, h);
    connLogSubConnFailed(h->connID, "tcp", "timeout");
    cbdataUnlock(h);
    cbdataFree(h);
    debug(17, 2) ("pathSelectHoldTimeout invoked on fd %d\n", fd);
    comm_close(fd);
}

static void
pathSelectHoldConnect(int fd, int status, void *data)
{
    conn_id_t connID;
    id_holder *h = data;
    
    comm_remove_close_handler(fd, pathSelectHoldClosed, h);
    
    getCurrentTime();
    connID = h->connID;
    cbdataUnlock(h);

    debug(17, 2) ("pathSelectHoldConnect invoked on fd %d\n", fd);
    comm_close(fd);
}

void
pathSelectSpecClosed(int fd, void *data)
{
    FwdState *fwdState = data;
    int i;
    
    debug(17, 2) ("pathSelectSpecClosed FD %d %s\n", fd, storeUrl(fwdState->entry));
    for (i = 0; i < MAX_SPEC_OPENS; i++) {
	if (fwdState->specOpens[i].fd == fd) break;
    }
    assert(i < MAX_SPEC_OPENS);
    fwdState->specOpens[i].fd = -1;
    fwdState->open_specs--;
    cbdataUnlock(fwdState);
}


static void
pathSelectStartICPConnects(FwdState *fwdState)
{
    int n;
    int n_sent;
    
    if (!cbdataValid(fwdState) ||
	(fwdState->connect_via != FWD_CONNECT_UNKNOWN) ||
	(fwdState->icp_retries >= MAX_ICP_RETRIES)) {
	/* If there are no direct attempts, then we're responsible
	 * for tracking down the failure.  If there _are_ direct
	 * attempts, then pathSelectSpecTimeout() will eventually
	 * get them and we can just return. */
	if (fwdState->connect_via == FWD_CONNECT_UNKNOWN &&
	    fwdState->use_peers == FWD_ALWAYS_PEERS) {
	    ErrorState *err;
	    err = errorCon(ERR_CONNECT_FAIL, HTTP_GATEWAY_TIMEOUT);
	    err->request = requestLink(fwdState->request);
	    err->xerrno = ETIMEDOUT;
	    fwdFail(fwdState, err);
	    fwdStateFree(fwdState);
	}
	cbdataUnlock(fwdState);
	return;
    }
    fwdState->icp_retries++;

    debug(17, 3) ("pathSelectStartICPConnects starting\n");
    /* There should be some cache digest / carp / netdb checks here */
    /* but there aren't. ;-) */

    if (!fwdState->request->flags.hierarchical &&
	fwdState->use_peers != FWD_ALWAYS_PEERS) {
	cbdataUnlock(fwdState);
	return;
    }

    if (EBIT_TEST(fwdState->entry->flags, KEY_PRIVATE) &&
	!neighbors_do_private_keys &&
	fwdState->use_peers != FWD_ALWAYS_PEERS) {
	cbdataUnlock(fwdState);
	return;
    }
    n = neighborsCount(fwdState->request);
    n_sent = neighborsUdpPing(fwdState->request,
			      fwdState->entry,
			      pathSelectHandlePingReply,
			      fwdState,
			      &(fwdState->ping.n_replies_expected),
			      &(fwdState->ping.timeout),
			      ICP_FLAG_SETUP,
			      fwdState->connID);
    if (n_sent > 0) fwdState->entry->ping_status = PING_WAITING;
    /* And deal with timeouts and such */
    debug(17, 3) ("sent %d neighbor pings\n", n_sent);
    
    eventAdd("ICPRetransmissions", pathSelectStartICPConnects, fwdState,
	     (double)3.0, 1);
}

static void
pathSelectHandlePingReply(peer * p, peer_t type, protocol_t proto, void *pingdata, void *data, struct sockaddr_in *to)
{
    if (proto == PROTO_ICP)
	pathSelectHandleICPReply(p, type, pingdata, data, to);
#if USE_HTCP
    else if (proto == PROTO_HTCP)
	pathSelectHandleHTCPReply(p, type, pingdata, data, to);
#endif
}

static void
pathSelectHandleDelayedIcpReply(void *data)
{
    /* Do something */
    /* then hand it back to 
     * with the flags munged... */
    PeerDelayed *peerDelayed = data;
    icp_common_t header;
    struct sockaddr_in to;

    debug(17, 3) ("pSHandleDelayedIcpReply starting\n");

    header = peerDelayed->header;
    header.flags &= ~ICP_FLAG_SETUP_PCONN;
    to = peerDelayed->to;
    
    if (cbdataValid(peerDelayed->data)) {
	pathSelectHandleICPReply(peerDelayed->p, peerDelayed->type,
				 &header, peerDelayed->data,
				 peerDelayed->has_to ? &to : NULL);
    }
    
    cbdataUnlock(peerDelayed->data);
    cbdataFree(peerDelayed);
}

static void
pathSelectHandleICPReply(peer * p, peer_t type, icp_common_t * header,
			 void *data, struct sockaddr_in *to)
{
    /* XXX:  This breaks parent selection. */
    FwdState *fwdState = data;
    icp_opcode op = header->opcode;
    char *status_name;

    if (fwdState->connect_via != FWD_CONNECT_UNKNOWN) {
	return;
    }

    if (op != ICP_HIT &&
	!(op == ICP_MISS && (header->flags & ICP_FLAG_SETUP))) {
	return;
    }

    if (op == ICP_HIT) { status_name = "HIT"; }
    else if (header->flags & ICP_FLAG_SETUP_PCONN) { status_name = "PCONN"; }
    else { status_name = "SPEC"; }

    debug(17, 3) ("pathSelectHandleICPReply %s for %s\n",
		  status_name,
		  storeUrl(fwdState->entry));

    if (op == ICP_MISS && (header->flags & ICP_FLAG_SETUP) &&
	header->flags & ICP_FLAG_SETUP_PCONN) {
	u_num32 rttms;
	PeerDelayed *peerDelayed;

	rttms = header->pad & 0xFFFF;
	debug(17, 3) ("pSHandleICPReply PCONN hit rtt %d ms for '%s'\n",
		      rttms, storeUrl(fwdState->entry));
	peerDelayed = cbdataAlloc(PeerDelayed);
	if (!peerDelayed) {
	    debug(17, 1) ("cbdataAlloc of PeerDelayed failed\n");
	    return;
	}
	peerDelayed->p = p;
	peerDelayed->type = type;
	peerDelayed->header = *header;
	peerDelayed->data = data;
	if (to != NULL) {
	    peerDelayed->to = *to;
	    peerDelayed->has_to = 1;
	} else {
	    peerDelayed->has_to = 0;
	}
	cbdataLock(peerDelayed->data);
	
	eventAdd("DelayedICPCB", pathSelectHandleDelayedIcpReply, peerDelayed,
		 (double)(rttms) / 1000, 1);

	return;
    }
    
    fwdState->connect_via = FWD_CONNECT_PEER;

    /* XXX:  Must annotate so that we know which outbound interface
     * to use - or contact servers in parallel... */
    peerAddFwdServer(&fwdState->servers, p, SIBLING_HIT);
    if (fwdState->use_peers != FWD_ALWAYS_PEERS)
	peerAddFwdServer(&fwdState->servers, NULL, DIRECT);
    
    fwdState->entry->ping_status = PING_DONE;
    {
	struct in_addr srca;
	bzero(&srca, sizeof(srca));
	
	connLogFetchViaPeer(fwdState->connID,
			    srca, 0,
			p->in_addr.sin_addr, p->in_addr.sin_port);
    }
    fwdConnectStart(fwdState);
}
#if USE_HTCP
static void
pathSelectHandleHTCPReply(peer * p, peer_t type, htcpReplyData * header,
			 void *data, struct sockaddr_in *to)
{
   icp_common_t icpheader;

   debug(17,3) ("Entering pathSelectHandleHTCPReply\n");

   icpheader.opcode = (header->hit == 1) ? ICP_HIT : ICP_MISS;
   icpheader.version = 0;
   icpheader.length = 0;
   icpheader.reqnum = header->msg_id;
   icpheader.flags = header->flags;
   icpheader.pad = (u_num32) header->cto.rtt;
   icpheader.shostid = header->shostid;

   pathSelectHandleICPReply(p, type, &icpheader, data, to);
}
#endif   

int
fwdCheckDeferRead(int fd, void *data)
{
    StoreEntry *e = data;
    MemObject *mem = e->mem_obj;
    int rc = 0;
    if (mem == NULL)
	return 0;
#if URL_CHECKSUM_DEBUG
    assert(e->mem_obj->chksum == url_checksum(e->mem_obj->url));
#endif
#if DELAY_POOLS
    if (fd < 0)
	(void) 0;
    else if (delayIsNoDelay(fd))
	(void) 0;
    else {
	int i = delayMostBytesWanted(mem, INT_MAX);
	if (0 == i)
	    return 1;
	/* was: rc = -(rc != INT_MAX); */
	else if (INT_MAX == i)
	    rc = 0;
	else
	    rc = -1;
    }
#endif
    if (EBIT_TEST(e->flags, ENTRY_FWD_HDR_WAIT))
	return rc;
    if (mem->inmem_hi - storeLowestMemReaderOffset(e) < READ_AHEAD_GAP)
	return rc;
    return 1;
}

void
fwdFail(FwdState * fwdState, ErrorState * errorState)
{
    assert(EBIT_TEST(fwdState->entry->flags, ENTRY_FWD_HDR_WAIT));
    debug(17, 3) ("fwdFail: %s \"%s\"\n\t%s\n",
	err_type_str[errorState->type],
	httpStatusString(errorState->http_status),
	storeUrl(fwdState->entry));
    if (fwdState->err)
	errorStateFree(fwdState->err);
    fwdState->err = errorState;
}

/*
 * Called when someone else calls StoreAbort() on this entry
 */
static void
fwdAbort(void *data)
{
    FwdState *fwdState = data;
    debug(17, 2) ("fwdAbort: %s\n", storeUrl(fwdState->entry));
    fwdStateFree(fwdState);
}

/*
 * Frees fwdState without closing FD or generating an abort
 */
void
fwdUnregister(int fd, FwdState * fwdState)
{
    debug(17, 3) ("fwdUnregister: %s\n", storeUrl(fwdState->entry));
    assert(fd == fwdState->server_fd);
    assert(fd > -1);
    comm_remove_close_handler(fd, fwdServerClosed, fwdState);
    fwdState->server_fd = -1;
}

/*
 * server-side modules call fwdComplete() when they are done
 * downloading an object.  Then, we either 1) re-forward the
 * request somewhere else if needed, or 2) call storeComplete()
 * to finish it off
 */
void
fwdComplete(FwdState * fwdState)
{
    StoreEntry *e = fwdState->entry;
    assert(e->store_status == STORE_PENDING);
    debug(17, 3) ("fwdComplete: %s\n\tstatus %d\n", storeUrl(e),
	e->mem_obj->reply->sline.status);
#if URL_CHECKSUM_DEBUG
    assert(e->mem_obj->chksum == url_checksum(e->mem_obj->url));
#endif
    fwdLogReplyStatus(fwdState->n_tries, e->mem_obj->reply->sline.status);
    if (fwdReforward(fwdState)) {
	debug(17, 3) ("fwdComplete: re-forwarding %d %s\n",
	    e->mem_obj->reply->sline.status,
	    storeUrl(e));
	if (fwdState->server_fd > -1)
	    fwdUnregister(fwdState->server_fd, fwdState);
	storeEntryReset(e);
	fwdStartComplete(fwdState->servers, fwdState);
    } else {
	debug(17, 3) ("fwdComplete: not re-forwarding status %d\n",
	    e->mem_obj->reply->sline.status);
	EBIT_CLR(e->flags, ENTRY_FWD_HDR_WAIT);
	storeComplete(e);
	/*
	 * If fwdState isn't associated with a server FD, it
	 * won't get freed unless we do it here.
	 */
	if (fwdState->server_fd < 0)
	    fwdStateFree(fwdState);
    }
}

void
fwdInit(void)
{
    cachemgrRegister("forward",
	"Request Forwarding Statistics",
	fwdStats, 0, 1);
#if WIP_FWD_LOG
    if (logfile)
	(void) 0;
    else if (NULL == Config.Log.forward)
	(void) 0;
    else
	logfile = logfileOpen(Config.Log.forward, 0, 1);
#endif
}

static void
fwdLogReplyStatus(int tries, http_status status)
{
    if (status > HTTP_INVALID_HEADER)
	return;
    assert(tries);
    tries--;
    if (tries > MAX_FWD_STATS_IDX)
	tries = MAX_FWD_STATS_IDX;
    FwdReplyCodes[tries][status]++;
}

static void
fwdStats(StoreEntry * s)
{
    int i;
    int j;
    storeAppendPrintf(s, "Status");
    for (j = 0; j <= MAX_FWD_STATS_IDX; j++) {
	storeAppendPrintf(s, "\ttry#%d", j + 1);
    }
    storeAppendPrintf(s, "\n");
    for (i = 0; i <= (int) HTTP_INVALID_HEADER; i++) {
	if (FwdReplyCodes[0][i] == 0)
	    continue;
	storeAppendPrintf(s, "%3d", i);
	for (j = 0; j <= MAX_FWD_STATS_IDX; j++) {
	    storeAppendPrintf(s, "\t%d", FwdReplyCodes[j][i]);
	}
	storeAppendPrintf(s, "\n");
    }
}

int
fwdReforwardableStatus(http_status s)
{
    switch (s) {
    case HTTP_FORBIDDEN:
    case HTTP_INTERNAL_SERVER_ERROR:
    case HTTP_NOT_IMPLEMENTED:
    case HTTP_BAD_GATEWAY:
    case HTTP_SERVICE_UNAVAILABLE:
    case HTTP_GATEWAY_TIMEOUT:
	return 1;
    default:
	return 0;
    }
    /* NOTREACHED */
}

#if WIP_FWD_LOG
void
fwdUninit(void)
{
    if (NULL == logfile)
	return;
    logfileClose(logfile);
    logfile = NULL;
}

void
fwdLogRotate(void)
{
    if (logfile)
	logfileRotate(logfile);
}

static void
fwdLog(FwdState * fwdState)
{
    if (NULL == logfile)
	return;
    logfilePrintf(logfile, "%9d.%03d %03d %s %s\n",
	(int) current_time.tv_sec,
	(int) current_time.tv_usec / 1000,
	fwdState->last_status,
	RequestMethodStr[fwdState->request->method],
	fwdState->request->canonical);
}

void
fwdStatus(FwdState * fwdState, http_status s)
{
    fwdState->last_status = s;
}

#endif
