
/*
 * $Id: htcp.c,v 1.5 2004/07/27 23:12:39 rohit Exp $
 *
 * DEBUG: section 31    Hypertext Caching Protocol
 * AUTHOR: Duane Wesssels
 *
 * SQUID Web Proxy Cache          http://www.squid-cache.org/
 * ----------------------------------------------------------
 *
 *  Squid is the result of efforts by numerous individuals from
 *  the Internet community; see the CONTRIBUTORS file for full
 *  details.   Many organizations have provided support for Squid's
 *  development; see the SPONSORS file for full details.  Squid is
 *  Copyrighted (C) 2001 by the Regents of the University of
 *  California; see the COPYRIGHT file for full details.  Squid
 *  incorporates software developed and/or copyrighted by other
 *  sources; see the CREDITS file for full details.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 *
 */



/* HTCP v0.1 protocol details:
 * 
 * On request, send RESPONSE=15 in order to open specs
 * In theory, random proxies will drop this malformed request
 * 
 * On reply, send RESPONSE=15 to signal cache headers included
 * Cache headers will include a Cache-RTT: entry.
 *
 * Use the monet option to indicate you're talking to a modified proxy
 */




#include "squid.h"

typedef struct _Countstr Countstr;
typedef struct _htcpHeader htcpHeader;
typedef struct _htcpDataHeader htcpDataHeader;
typedef struct _htcpAuthHeader htcpAuthHeader;
typedef struct _htcpStuff htcpStuff;
typedef struct _htcpSpecifier htcpSpecifier;
typedef struct _htcpDetail htcpDetail;

struct _Countstr {
    u_short length;
    char *text;
};

struct _htcpHeader {
    u_short length;
    u_char major;
    u_char minor;
};

struct _htcpDataHeader {
    u_short length;
#if !WORDS_BIGENDIAN
    unsigned int opcode:4;
    unsigned int response:4;
#else
    unsigned int response:4;
    unsigned int opcode:4;
#endif
#if !WORDS_BIGENDIAN
    unsigned int reserved:6;
    unsigned int F1:1;
    unsigned int RR:1;
#else
    unsigned int RR:1;
    unsigned int F1:1;
    unsigned int reserved:6;
#endif
    u_num32 msg_id;
};

    /* RR == 0 --> F1 = RESPONSE DESIRED FLAG */
    /* RR == 1 --> F1 = MESSAGE OVERALL FLAG */
    /* RR == 0 --> REQUEST */
    /* RR == 1 --> RESPONSE */

struct _htcpAuthHeader {
    u_short length;
    time_t sig_time;
    time_t sig_expire;
};

struct _htcpSpecifier {
    char *method;
    char *uri;
    char *version;
    char *req_hdrs;
};

struct _htcpDetail {
    char *resp_hdrs;
    char *entity_hdrs;
    char *cache_hdrs;
};

struct _htcpStuff {
    int op;
    int rr;
    int f1;
    int response;
    u_num32 msg_id;
    htcpSpecifier S;
    htcpDetail D;
};

#define MONET_FLAG 15

enum {
    HTCP_NOP,
    HTCP_TST,
    HTCP_MON,
    HTCP_SET,
    HTCP_CLR,
    HTCP_END
};

static const char *const htcpOpcodeStr[] =
{
    "HTCP_NOP",
    "HTCP_TST",
    "HTCP_MON",
    "HTCP_SET",
    "HTCP_CLR",
    "HTCP_END"
};

/*
 * values for htcpDataHeader->response
 */
enum {
    AUTH_REQUIRED,
    AUTH_FAILURE,
    OPCODE_UNIMPLEMENTED,
    MAJOR_VERSION_UNSUPPORTED,
    MINOR_VERSION_UNSUPPORTED,
    INVALID_OPCODE
};

/*
 * values for htcpDataHeader->RR
 */
enum {
    RR_REQUEST,
    RR_RESPONSE
};


/* structures used for multisend()
 */

#define MAX_HTCP_ADDRS 10

struct HtcpFDMap {
  struct sockaddr_in sa;
  int fd;
};

#define MAX_HTCP_SOCK 100
struct sockaddr_in htcp_fdinfos[MAX_HTCP_SOCK];
int n_htcp_fdaddrs;
struct HtcpFDMap htcp_fdaddrs[MAX_HTCP_ADDRS];

static u_num32 msg_id_counter = 0;
static int htcpInSocket = -1;
static int htcpOutSocket = -1;
#define N_QUERIED_KEYS 256
static cache_key queried_keys[N_QUERIED_KEYS][MD5_DIGEST_CHARS];
static MemPool *htcpSpecifierPool = NULL;
static MemPool *htcpDetailPool = NULL;


static char *htcpBuildPacket(htcpStuff * stuff, ssize_t * len);
static htcpSpecifier *htcpUnpackSpecifier(char *buf, int sz);
static htcpDetail *htcpUnpackDetail(char *buf, int sz);
static int htcpUnpackCountstr(char *buf, int sz, char **str);
static ssize_t htcpBuildEmptyAuth(char *buf, size_t buflen);
static ssize_t htcpBuildCountstr(char *buf, size_t buflen, const char *s);
static ssize_t htcpBuildSig(char *buf, size_t buflen, const unsigned char *s);
static ssize_t htcpBuildData(char *buf, size_t buflen, htcpStuff * stuff);
static ssize_t htcpBuildDetail(char *buf, size_t buflen, htcpStuff * stuff);
static ssize_t htcpBuildOpData(char *buf, size_t buflen, htcpStuff * stuff);
static ssize_t htcpBuildSpecifier(char *buf, size_t buflen, htcpStuff * stuff);
static ssize_t htcpBuildTstOpData(char *buf, size_t buflen, htcpStuff * stuff);
static void htcpFreeSpecifier(htcpSpecifier * s);
static void htcpFreeDetail(htcpDetail * s);
static void htcpHandle(int fd, char *buf, int sz, struct sockaddr_in *from);
static void htcpHandleData(int fd, char *buf, int sz, struct sockaddr_in *from);
static void htcpHandleMon(htcpDataHeader *, char *buf, int sz, struct sockaddr_in *from);
static void htcpHandleNop(htcpDataHeader *, char *buf, int sz, struct sockaddr_in *from);
static void htcpHandleSet(htcpDataHeader *, char *buf, int sz, struct sockaddr_in *from);
static void htcpHandleTst(int fd, htcpDataHeader *, char *buf, int sz, struct sockaddr_in *from, int servid);
static void htcpRecv(int fd, void *data);
static void htcpSend(const char *buf, int len, struct sockaddr_in *to, conn_id_t connID, htcpStuff *stuff);
static void htcpMyAddrlistSend(char * buf, int buflen, struct sockaddr_in *to, conn_id_t ConnID, htcpStuff *stuff);
static void htcpTstReply(int, htcpDataHeader *, StoreEntry *, htcpSpecifier *, struct sockaddr_in *, conn_id_t);
static void htcpHandleTstRequest(int fd, htcpDataHeader *, char *buf, int sz, struct sockaddr_in *from, int servid);
static void htcpHandleTstResponse(int, htcpDataHeader *, char *, int, struct sockaddr_in *, int);
static StoreEntry *htcpCheckHit(const htcpSpecifier *);



const cache_key *
hmac_md5(char* text, int text_len, char * key, int key_len)
{
        MD5_CTX context;
	static cache_key digest[MD5_DIGEST_CHARS];

        unsigned char k_ipad[65];    /* inner padding -
                                      * key XORd with ipad
                                      */
        unsigned char k_opad[65];    /* outer padding -
                                      * key XORd with opad
                                      */
        unsigned char tk[16];
        int i;
        /* if key is longer than 64 bytes reset it to key=MD5(key) */
        if (key_len > 64) {

                MD5_CTX      tctx;

                MD5Init(&tctx);
                MD5Update(&tctx, key, key_len);
                MD5Final(tk, &tctx);

                key = tk;
                key_len = 16;
        }

        /*
         * the HMAC_MD5 transform looks like:
         *
         * MD5(K XOR opad, MD5(K XOR ipad, text))
         *
         * where K is an n byte key
         * ipad is the byte 0x36 repeated 64 times

         * opad is the byte 0x5c repeated 64 times
         * and text is the data being protected
         */

        /* start out by storing key in pads */
        bzero( k_ipad, sizeof k_ipad);
        bzero( k_opad, sizeof k_opad);
        bcopy( key, k_ipad, key_len);
        bcopy( key, k_opad, key_len);

        /* XOR key with ipad and opad values */
        for (i=0; i<64; i++) {
                k_ipad[i] ^= 0x36;
                k_opad[i] ^= 0x5c;
        }
        /*
         * perform inner MD5
         */
        MD5Init(&context);                   /* init context for 1st
                                              * pass */
        MD5Update(&context, k_ipad, 64);      /* start with inner pad */
        MD5Update(&context, text, text_len); /* then text of datagram */
        MD5Final(digest, &context);          /* finish up 1st pass */
        /*
         * perform outer MD5
         */
        MD5Init(&context);                   /* init context for 2nd
                                              * pass */
        MD5Update(&context, k_opad, 64);     /* start with outer pad */
        MD5Update(&context, digest, 16);     /* then results of 1st
                                              * hash */
        MD5Final(digest, &context);          /* finish up 2nd pass */

	return digest;
}


static void
htcpHexdump(const char *tag, const char *s, int sz)
{
#if USE_HEXDUMP
    int i;
    int k;
    char hex[80];
    debug(31, 3) ("htcpHexdump %s\n", tag);
    memset(hex, '\0', 80);
    for (i = 0; i < sz; i++) {
	k = i % 16;
	snprintf(&hex[k * 3], 4, " %02x", (int) *(s + i));
	if (k < 15 && i < (sz - 1))
	    continue;
	debug(31, 3) ("\t%s\n", hex);
	memset(hex, '\0', 80);
    }
#endif
}

/*
 * STUFF FOR SENDING HTCP MESSAGES
 */

static ssize_t
htcpBuildEmptyAuth(char *buf, size_t buflen)
{
    htcpAuthHeader auth;
    size_t copy_sz = 0;
    assert(2 == sizeof(u_short));
    auth.length = htons(2);
    copy_sz += 2;
    assert(buflen >= copy_sz);
    xmemcpy(buf, &auth, copy_sz);
    return copy_sz;
}



static ssize_t
htcpBuildAuth(char *buf, size_t buflen, char *data, size_t dlen)
{
  /* Basic (somewhat inefficient) idea here is to simultaneously
   * build both the auth entry and a separate buffer which holds
   * the stuff we need to MD5.  Then we do the HMAC and write the
   * signature.
   */
  
    htcpAuthHeader auth;
    size_t copy_sz = 0;
    off_t off = 0;
    u_short offinv;
    ssize_t s;
    const cache_key *digest;
    static char authbuf[8192];
    int authoff = 0;

    memset(&authbuf, '\0', 8192);

    /*
    // First up we have src addr + port
    // These are already in network order!
    debug(31,1) ("Src addr is: %s\n", inet_ntoa(f->sa.sin_addr));
    debug(31,1) ("Src port is: %d\n", ntohs(f->sa.sin_port));

    debug(31,1) ("1Current location: %p\n", authbuf + authoff);
    xmemcpy(authbuf, &(f->sa.sin_addr), 4);
    authoff += 4;
    debug(31,1) ("2Current location: %p\n", authbuf + authoff);

    xmemcpy(authbuf + authoff, &(f->sa.sin_port), 2);
    authoff += 2;
    debug(31,1) ("3Current location: %p\n", authbuf + authoff);


    // Then we do dst addr + port
    // These are already in network order!
    debug(31,1) ("Dst addr is: %s\n", inet_ntoa(p->in_addr.sin_addr));
    debug(31,1) ("Dst port is: %d\n", ntohs(p->in_addr.sin_port));

    xmemcpy(authbuf + authoff, &(p->in_addr.sin_addr), 4);
    authoff += 4;
    debug(31,1) ("4Current location: %p\n", authbuf + authoff);

    xmemcpy(authbuf + authoff, &(p->in_addr.sin_port), 2);
    authoff += 2;
    debug(31,1) ("5Current location: %p\n", authbuf + authoff);

    // Next comes the HTCP version number.  This is going to be 2 bytes of \0
    authoff += 2;

*/

    // NATs will make these meaningless.  Leave the src and dst addrs = 0
    // Also, HTCP major/minor = 0/0
    authoff += 14;


    assert(2 == sizeof(u_short));

    // Now we set up the times
    getCurrentTime();
    auth.sig_time = squid_curtime;
    auth.sig_expire = auth.sig_time + 60;

    assert(4 == sizeof(time_t));
    auth.sig_time = htonl(auth.sig_time);
    auth.sig_expire = htonl(auth.sig_expire);
    copy_sz = sizeof(htcpAuthHeader);

    // Copy this part into the buffer...
    assert(buflen >= copy_sz);
    xmemcpy(buf, &auth, copy_sz);
    off += copy_sz;

    // And also the times go into the authentication buffer...
    xmemcpy(authbuf + authoff, &auth.sig_time, 4);
    authoff += 4;
    xmemcpy(authbuf + authoff, &auth.sig_expire, 4);
    authoff += 4;

    // Next comes the HTCP OP data
    xmemcpy(authbuf + authoff, data, dlen);
    authoff += dlen;
    //fprintf(stderr, "%d\n", dlen);

    //Next we do the keyname.  For now, we leave it blank.
    s = htcpBuildCountstr(buf + off, buflen - off, HTCPSenderID);
    if (s < 0)
	return s;
    off += s;

    s = htcpBuildCountstr(authbuf + authoff, 8192 - authoff, HTCPSenderID);
    if (s < 0)
	return s;
    authoff += s; 
    //fprintf(stderr, "%d\n", s);

    // Now we can do the HMAC!
    debug(31,7) ("htcpBuildAuth: Adding signature\n");
    digest = hmac_md5(Config.htcpkey, strlen(Config.htcpkey), authbuf, authoff);



    // Finally, we build the signature.  Somehow.
    s = htcpBuildSig(buf + off, buflen - off, digest);
    if (s < 0)
	return s;
    off += s;

    // Fill in the length
    offinv = htons(off);
    xmemcpy(buf, &offinv, 2);

    return off;
}

static ssize_t
htcpBuildCountstr(char *buf, size_t buflen, const char *s)
{
    u_short length;
    size_t len;
    off_t off = 0;
    if (buflen - off < 2)
	return -1;
    if (s)
	len = strlen(s);
    else
	len = 0;
    debug(31, 3) ("htcpBuildCountstr: LENGTH = %d\n", len);
    debug(31, 3) ("htcpBuildCountstr: TEXT = {%s}\n", s ? s : "<NULL>");
    length = htons((u_short) len);
    xmemcpy(buf + off, &length, 2);
    off += 2;
    if (buflen - off < len)
	return -1;
    if (len)
	xmemcpy(buf + off, s, len);
    off += len;
    return off;
}

static ssize_t
htcpBuildSig(char *buf, size_t buflen, const unsigned char *s)
{
    u_short length;
    size_t len;
    off_t off = 0;
    if (buflen - off < 2)
	return -1;
    len = MD5_DIGEST_CHARS;
    debug(31, 3) ("htcpBuildSig: LENGTH = %d\n", len);
    length = htons((u_short) len);
    xmemcpy(buf + off, &length, 2);
    off += 2;
    if (buflen - off < len)
	return -1;
    if (len)
	xmemcpy(buf + off, s, len);
    off += len;
    return off;
}

static ssize_t
htcpBuildSpecifier(char *buf, size_t buflen, htcpStuff * stuff)
{
    ssize_t off = 0;
    ssize_t s;
    s = htcpBuildCountstr(buf + off, buflen - off, stuff->S.method);
    if (s < 0)
	return s;
    off += s;
    debug(31, 2) ("htcpBuildSpecifier: URI = %s\n", stuff->S.uri);
    s = htcpBuildCountstr(buf + off, buflen - off, stuff->S.uri);
    if (s < 0)
	return s;
    off += s;
    s = htcpBuildCountstr(buf + off, buflen - off, stuff->S.version);
    if (s < 0)
	return s;
    off += s;
    s = htcpBuildCountstr(buf + off, buflen - off, stuff->S.req_hdrs);
    if (s < 0)
	return s;
    off += s;
    debug(31, 3) ("htcpBuildSpecifier: size %d\n", (int) off);
    return off;
}

static ssize_t
htcpBuildDetail(char *buf, size_t buflen, htcpStuff * stuff)
{
    ssize_t off = 0;
    ssize_t s;
    s = htcpBuildCountstr(buf + off, buflen - off, stuff->D.resp_hdrs);
    if (s < 0)
	return s;
    off += s;
    s = htcpBuildCountstr(buf + off, buflen - off, stuff->D.entity_hdrs);
    if (s < 0)
	return s;
    off += s;
    s = htcpBuildCountstr(buf + off, buflen - off, stuff->D.cache_hdrs);
    if (s < 0)
	return s;
    off += s;
    return off;
}

static ssize_t
htcpBuildTstOpData(char *buf, size_t buflen, htcpStuff * stuff)
{
    switch (stuff->rr) {
    case RR_REQUEST:
	debug(31, 3) ("htcpBuildTstOpData: RR_REQUEST\n");
	return htcpBuildSpecifier(buf, buflen, stuff);
    case RR_RESPONSE:
	debug(31, 3) ("htcpBuildTstOpData: RR_RESPONSE\n");
	debug(31, 3) ("htcpBuildTstOpData: F1 = %d\n", stuff->f1);
	if (stuff->f1)		/* cache miss */
	    return 0;
	else			/* cache hit */
	    return htcpBuildDetail(buf, buflen, stuff);
    default:
	fatal_dump("htcpBuildTstOpData: bad RR value");
    }
    return 0;
}

static ssize_t
htcpBuildOpData(char *buf, size_t buflen, htcpStuff * stuff)
{
    ssize_t off = 0;
    debug(31, 3) ("htcpBuildOpData: opcode %s\n",
	htcpOpcodeStr[stuff->op]);
    switch (stuff->op) {
    case HTCP_TST:
	off = htcpBuildTstOpData(buf + off, buflen, stuff);
	break;
    default:
	assert(0);
	break;
    }
    return off;
}

static ssize_t
htcpBuildData(char *buf, size_t buflen, htcpStuff * stuff)
{
    ssize_t off = 0;
    ssize_t op_data_sz;
    size_t hdr_sz = sizeof(htcpDataHeader);
    htcpDataHeader hdr;
    if (buflen < hdr_sz)
	return -1;
    off += hdr_sz;		/* skip! */
    op_data_sz = htcpBuildOpData(buf + off, buflen - off, stuff);
    if (op_data_sz < 0)
	return op_data_sz;
    off += op_data_sz;
    debug(31, 3) ("htcpBuildData: hdr.length = %d\n", (int) off);
    hdr.length = (u_short) off;
    hdr.opcode = stuff->op;
    hdr.response = stuff->response;
    hdr.RR = stuff->rr;
    hdr.F1 = stuff->f1;
    hdr.msg_id = stuff->msg_id;
    /* convert multi-byte fields */
    hdr.length = htons(hdr.length);
    hdr.msg_id = htonl(hdr.msg_id);
    xmemcpy(buf, &hdr, hdr_sz);
    debug(31, 3) ("htcpBuildData: size %d\n", (int) off);
    return off;
}

static char *
htcpBuildPacket(htcpStuff * stuff, ssize_t * len)
{
    size_t buflen = 8192;
    ssize_t s;
    ssize_t off = 0;
    size_t hdr_sz = sizeof(htcpHeader);
    htcpHeader hdr;
    char *buf = xcalloc(buflen, 1);
    /* skip the header -- we don't know the overall length */
    if (buflen < hdr_sz) {
	xfree(buf);
	return NULL;
    }
    off += hdr_sz;
    s = htcpBuildData(buf + off, buflen - off, stuff);
    if (s < 0) {
	xfree(buf);
	return NULL;
    }
    off += s;
    s = htcpBuildEmptyAuth(buf + off, buflen - off);
    if (s < 0) {
	xfree(buf);
	return NULL;
    }
    off += s;
    hdr.length = htons((u_short) off);
    hdr.major = 0;
    hdr.minor = 0;
    xmemcpy(buf, &hdr, hdr_sz);
    *len = off;
    debug(31, 3) ("htcpBuildPacket: size %d\n", (int) off);
    return buf;
}

static char *
htcpBuildMonetPacket(htcpStuff * stuff, ssize_t * len)
{
    size_t buflen = 8192;
    ssize_t s;
    ssize_t a;
    ssize_t off = 0;
    size_t hdr_sz = sizeof(htcpHeader);
    htcpHeader hdr;
    char *buf = xcalloc(buflen, 1);
    /* skip the header -- we don't know the overall length */
    if (buflen < hdr_sz) {
	xfree(buf);
	return NULL;
    }
    off += hdr_sz;
    s = htcpBuildData(buf + off, buflen - off, stuff);
    if (s < 0) {
	xfree(buf);
	return NULL;
    }

    a = htcpBuildAuth(buf + off + s, buflen - off - s, buf + off, s);
    if (a < 0) {
	xfree(buf);
	return NULL;
    }
    off += s;
    off += a;
    hdr.length = htons((u_short) off);
    hdr.major = 0;
    hdr.minor = 0;
    xmemcpy(buf, &hdr, hdr_sz);
    *len = off;
    debug(31, 3) ("htcpBuildMonetPacket: size %d\n", (int) off);
    return buf;
}

static void
htcpSend(const char *buf, int len, struct sockaddr_in *to, conn_id_t connID, htcpStuff *stuff)
{
    int x;
    struct sockaddr_in myaddr;

    bzero(&myaddr, sizeof(myaddr));
    if (htcpOutSocket < MAX_HTCP_SOCK) {
      myaddr = htcp_fdinfos[htcpOutSocket];
    }


    debug(31, 3) ("htcpSend: %s/%d\n",
	inet_ntoa(to->sin_addr), (int) ntohs(to->sin_port));
    htcpHexdump("htcpSend", buf, len);
    x = comm_udp_sendto(htcpOutSocket,
	to,
	sizeof(struct sockaddr_in),
	buf,
	len);
    if (x < 0)
	debug(31, 0) ("htcpSend: FD %d sendto: %s\n", htcpOutSocket, xstrerror());
    else {
	connLogSendICP(connID, myaddr.sin_addr, myaddr.sin_port,
		       to->sin_addr, to->sin_port,
		       0, ICPSenderID, stuff->msg_id,
		       (stuff->response == 15) ? ICP_FLAG_SETUP : 0, 0);
    }
}

static void
htcpSendq(const char *buf, int len, struct sockaddr_in *to, htcpStuff *stuff)
{
    int x;
    debug(31, 3) ("htcpSend: %s/%d\n",
	inet_ntoa(to->sin_addr), (int) ntohs(to->sin_port));
    htcpHexdump("htcpSend", buf, len);
    x = comm_udp_sendto(htcpOutSocket,
	to,
	sizeof(struct sockaddr_in),
	buf,
	len);
    if (x < 0)
	debug(31, 0) ("htcpSend: FD %d sendto: %s\n", htcpOutSocket, xstrerror());
}


static void
htcpSendFrom(const char *buf, int len, struct sockaddr_in *to, int from, conn_id_t connID, htcpStuff *stuff)
{
    int x;
    struct sockaddr_in myaddr;

    bzero(&myaddr, sizeof(myaddr));
    if (from < MAX_HTCP_SOCK) {
      myaddr = htcp_fdinfos[from];
    }

    debug(31, 3) ("htcpSendFrom: from FD %d to %s/%d\n",
    from, inet_ntoa(to->sin_addr), (int) ntohs(to->sin_port));
    htcpHexdump("htcpSendFrom", buf, len);
    x = comm_udp_sendto(from,
	to,
	sizeof(struct sockaddr_in),
	buf,
	len);
    if (x < 0)
	debug(31, 0) ("htcpSendFrom: FD %d sendto: %s\n", htcpOutSocket, xstrerror());
    else {
	connLogSendICP(connID, myaddr.sin_addr, myaddr.sin_port,
		       to->sin_addr, to->sin_port,
		       0, ICPSenderID, stuff->msg_id,
		       (stuff->response == 15) ? ICP_FLAG_SETUP : 0, 0);
    }
}


int
htcpMultiSend(char * buf, int buflen, int n_to_send, int *from, struct sockaddr_in *to, conn_id_t ConnID, htcpStuff *stuff) 
{
    int i;
    int n_sent = 0;
    struct sockaddr_in myaddr;
    int x;

    // I'm given a FD (from[i]).  Gotta lookup the corresponding IP/port (myaddr)

    for (i = 0; i < n_to_send; i++) {
      bzero(&myaddr, sizeof(myaddr));
      if (from[i] < MAX_HTCP_SOCK) {
          myaddr = htcp_fdinfos[from[i]];
      }

      debug(31, 3) ("htcpMultiSend: from FD %d to %s/%d\n",
		    from[i], inet_ntoa(to[i].sin_addr), (int) ntohs(to[i].sin_port));
      htcpHexdump("htcpMultiSend", buf, buflen);
      x = comm_udp_sendto(from[i],
			  &to[i],
			  sizeof(struct sockaddr_in),
			  buf,
			  buflen);
      if (x < 0)
	debug(31, 0) ("htcpMultiSend: FD %d sendto: %s\n", htcpOutSocket, xstrerror());
      else {
	n_sent++;

	connLogSendICP(ConnID, myaddr.sin_addr, myaddr.sin_port,
		       to[i].sin_addr, to[i].sin_port,
		       0, ICPSenderID, stuff->msg_id,
		       (stuff->response == 15) ? ICP_FLAG_SETUP : 0, 0);
      }
    }
    
    return n_sent++;
}


static void
htcpMyAddrlistSend(char * buf, int buflen, struct sockaddr_in *to, conn_id_t ConnID, htcpStuff *stuff)
{
    struct sockaddr_in saddrs[ICP_MAX_PAR_REPLIES];
    int fdarray[ICP_MAX_PAR_REPLIES];
    int i;

    memset(saddrs, '\0', sizeof(saddrs));

    if (n_htcp_fdaddrs > 0) {
	for (i = 0; i < n_htcp_fdaddrs; i++) {
	    saddrs[i] = *to;
	    fdarray[i] = htcp_fdaddrs[i].fd;
	}
	htcpMultiSend(buf, buflen, n_htcp_fdaddrs, fdarray, saddrs, ConnID, stuff);
    } else {
      htcpSend(buf, buflen, to, ConnID, stuff);
    }

}




/*
 * STUFF FOR RECEIVING HTCP MESSAGES
 */

static void
htcpFreeSpecifier(htcpSpecifier * s)
{
    safe_free(s->method);
    safe_free(s->uri);
    safe_free(s->version);
    safe_free(s->req_hdrs);
    memPoolFree(htcpSpecifierPool, s);
}

static void
htcpFreeDetail(htcpDetail * d)
{
    safe_free(d->resp_hdrs);
    safe_free(d->entity_hdrs);
    safe_free(d->cache_hdrs);
    memPoolFree(htcpDetailPool, d);
}

static int
htcpUnpackCountstr(char *buf, int sz, char **str)
{
    u_short l;
    debug(31, 3) ("htcpUnpackCountstr: sz = %d\n", sz);
    if (sz < 2) {
	debug(31, 3) ("htcpUnpackCountstr: sz < 2\n");
	return -1;
    }
    htcpHexdump("htcpUnpackCountstr", buf, sz);
    xmemcpy(&l, buf, 2);
    l = ntohs(l);
    buf += 2;
    sz -= 2;
    debug(31, 3) ("htcpUnpackCountstr: LENGTH = %d\n", (int) l);
    if (sz < l) {
	debug(31, 3) ("htcpUnpackCountstr: sz(%d) < l(%d)\n", sz, l);
	return -1;
    }
    if (str) {
	*str = xmalloc(l + 1);
	xstrncpy(*str, buf, l + 1);
	debug(31, 3) ("htcpUnpackCountstr: TEXT = {%s}\n", *str);
    }
    return (int) l + 2;
}

static int
htcpUnpackSig(char *buf, int sz, unsigned char **str)
{
    u_short l;
    debug(31, 3) ("htcpUnpackSig: sz = %d\n", sz);
    if (sz < 2) {
	debug(31, 3) ("htcpUnpackSig: sz < 2\n");
	return -1;
    }
    htcpHexdump("htcpUnpackSig", buf, sz);
    xmemcpy(&l, buf, 2);
    l = ntohs(l);
    buf += 2;
    sz -= 2;
    if ((l != MD5_DIGEST_CHARS) || (sz < l)) {
	debug(31, 3) ("htcpUnpackSig: sz(%d) < l(%d)!= MD5_DIGEST_CHARS\n", sz, l);
	return -1;
    }
    if (str) {
	*str = xmalloc(l);
	memcpy(*str, buf, MD5_DIGEST_CHARS);
    }
    return (int) l + 2;
}

static htcpSpecifier *
htcpUnpackSpecifier(char *buf, int sz)
{
    htcpSpecifier *s = memPoolAlloc(htcpSpecifierPool);
    int o;
    debug(31, 3) ("htcpUnpackSpecifier: %d bytes\n", (int) sz);
    o = htcpUnpackCountstr(buf, sz, &s->method);
    if (o < 0) {
	debug(31, 1) ("htcpUnpackSpecifier: failed to unpack METHOD\n");
	htcpFreeSpecifier(s);
	return NULL;
    }
    buf += o;
    sz -= o;
    o = htcpUnpackCountstr(buf, sz, &s->uri);
    if (o < 0) {
	debug(31, 1) ("htcpUnpackSpecifier: failed to unpack URI\n");
	htcpFreeSpecifier(s);
	return NULL;
    }
    buf += o;
    sz -= o;
    o = htcpUnpackCountstr(buf, sz, &s->version);
    if (o < 0) {
	debug(31, 1) ("htcpUnpackSpecifier: failed to unpack VERSION\n");
	htcpFreeSpecifier(s);
	return NULL;
    }
    buf += o;
    sz -= o;
    o = htcpUnpackCountstr(buf, sz, &s->req_hdrs);
    if (o < 0) {
	debug(31, 1) ("htcpUnpackSpecifier: failed to unpack REQ-HDRS\n");
	htcpFreeSpecifier(s);
	return NULL;
    }
    buf += o;
    sz -= o;
    debug(31, 3) ("htcpUnpackSpecifier: %d bytes left\n", sz);
    return s;
}

static htcpDetail *
htcpUnpackDetail(char *buf, int sz)
{
    htcpDetail *d = memPoolAlloc(htcpDetailPool);
    int o;
    debug(31, 3) ("htcpUnpackDetail: %d bytes\n", (int) sz);
    o = htcpUnpackCountstr(buf, sz, &d->resp_hdrs);
    if (o < 0) {
	debug(31, 1) ("htcpUnpackDetail: failed to unpack RESP_HDRS\n");
	htcpFreeDetail(d);
	return NULL;
    }
    buf += o;
    sz -= o;
    o = htcpUnpackCountstr(buf, sz, &d->entity_hdrs);
    if (o < 0) {
	debug(31, 1) ("htcpUnpackDetail: failed to unpack ENTITY_HDRS\n");
	htcpFreeDetail(d);
	return NULL;
    }
    buf += o;
    sz -= o;
    o = htcpUnpackCountstr(buf, sz, &d->cache_hdrs);
    if (o < 0) {
	debug(31, 1) ("htcpUnpackDetail: failed to unpack CACHE_HDRS\n");
	htcpFreeDetail(d);
	return NULL;
    }
    buf += o;
    sz -= o;
    debug(31, 3) ("htcpUnpackDetail: %d bytes left\n", sz);
    return d;
}


static void
htcpTstReply(int fd, htcpDataHeader *dhdr, StoreEntry * e, htcpSpecifier * spec, struct sockaddr_in *from, conn_id_t connID)
{
    htcpStuff stuff;
    char *pkt;
    HttpHeader hdr;
    MemBuf mb;
    Packer p;
    ssize_t pktlen;
    char *host;
    int rtt = 0;
    int hops = 0;
    int samp = 0;
    char cto_buf[128];
    int n_sent = 0;
    int i;

    memset(&stuff, '\0', sizeof(stuff));
    stuff.op = HTCP_TST;
    stuff.rr = RR_RESPONSE;
    stuff.f1 = 0;
    stuff.response = e ? 0 : 1;
    debug(31, 3) ("htcpTstReply: response = %d\n", stuff.response);
    stuff.msg_id = dhdr->msg_id;
    if (spec) {
	memBufDefInit(&mb);
	packerToMemInit(&p, &mb);
	httpHeaderInit(&hdr, hoHtcpReply);
	stuff.S.method = spec->method;
	stuff.S.uri = spec->uri;
	stuff.S.version = spec->version;
	stuff.S.req_hdrs = spec->req_hdrs;
	httpHeaderPutInt(&hdr, HDR_AGE,
	    e->timestamp <= squid_curtime ?
	    squid_curtime - e->timestamp : 0);
	httpHeaderPackInto(&hdr, &p);
	stuff.D.resp_hdrs = xstrdup(mb.buf);
	debug(31, 3) ("htcpTstReply: resp_hdrs = {%s}\n", stuff.D.resp_hdrs);
	memBufReset(&mb);
	httpHeaderReset(&hdr);
	if (e->expires > -1)
	    httpHeaderPutTime(&hdr, HDR_EXPIRES, e->expires);
	if (e->lastmod > -1)
	    httpHeaderPutTime(&hdr, HDR_LAST_MODIFIED, e->lastmod);
	httpHeaderPackInto(&hdr, &p);
	stuff.D.entity_hdrs = xstrdup(mb.buf);
	debug(31, 3) ("htcpTstReply: entity_hdrs = {%s}\n", stuff.D.entity_hdrs);
	memBufReset(&mb);
	httpHeaderReset(&hdr);
	if ((host = urlHostname(spec->uri))) {
	    netdbHostData(host, &samp, &rtt, &hops);
	    if (rtt || hops) {
		snprintf(cto_buf, 128, "%s %d %f %d",
		    host, samp, 0.001 * rtt, hops);
		httpHeaderPutExt(&hdr, "Cache-to-Origin", cto_buf);
	    }
	}
	httpHeaderPackInto(&hdr, &p);
	stuff.D.cache_hdrs = xstrdup(mb.buf);
	debug(31, 3) ("htcpTstReply: cache_hdrs = {%s}\n", stuff.D.cache_hdrs);
	memBufClean(&mb);
	httpHeaderClean(&hdr);
	packerClean(&p);
    }
    if (Config.htcpkey) {
      pkt = htcpBuildMonetPacket(&stuff, &pktlen);
    }
    else {
      pkt = htcpBuildPacket(&stuff, &pktlen);
    }

    if (pkt == NULL) {
	debug(31, 0) ("htcpTstReply: htcpBuildPacket() failed\n");
	return;
    }

    htcpSendFrom(pkt, (int) pktlen, from, fd, connID, &stuff);
    debug(31, 3) ("htcpTstReply: sent through default method\n");

    xfree(pkt);
}

static void
htcpHandleNop(htcpDataHeader * hdr, char *buf, int sz, struct sockaddr_in *from)
{
    debug(31, 3) ("htcpHandleNop: Unimplemented\n");
}

static StoreEntry *
htcpCheckHit(const htcpSpecifier * s)
{
    request_t *request;
    method_t m = urlParseMethod(s->method);
    StoreEntry *e = NULL, *hit = NULL;
    char *blk_end;
    request = urlParse(m, s->uri);
    if (NULL == request) {
	debug(31, 3) ("htcpCheckHit: NO; failed to parse URL\n");
	return NULL;
    }
    blk_end = s->req_hdrs + strlen(s->req_hdrs);
    if (!httpHeaderParse(&request->header, s->req_hdrs, blk_end)) {
	debug(31, 3) ("htcpCheckHit: NO; failed to parse request headers\n");
	goto miss;
    }
    e = storeGetPublicByRequest(request);
    if (NULL == e) {
	debug(31, 3) ("htcpCheckHit: NO; public object not found\n");
	goto miss;
    }
    if (!storeEntryValidToSend(e)) {
	debug(31, 3) ("htcpCheckHit: NO; entry not valid to send\n");
	goto miss;
    }
    if (refreshCheckHTCP(e, request)) {
	debug(31, 3) ("htcpCheckHit: NO; cached response is stale\n");
	goto miss;
    }
    debug(31, 3) ("htcpCheckHit: YES!?\n");
    hit = e;
  miss:
    requestDestroy(request);
    return hit;
}

static void
htcpHandleTst(int fd, htcpDataHeader * hdr, char *buf, int sz, struct sockaddr_in *from, int servid)
{
    debug(31, 3) ("htcpHandleTst: sz = %d\n", (int) sz);

    if (hdr->RR == RR_REQUEST)
	htcpHandleTstRequest(fd, hdr, buf, sz, from, servid);
    else
	htcpHandleTstResponse(fd, hdr, buf, sz, from, servid);
}

static void
htcpHandleTstResponse(int fd, htcpDataHeader * hdr, char *buf, int sz, struct sockaddr_in *from, int servid)
{
    htcpReplyData htcpReply;
    cache_key *key = NULL;
    htcpDetail *d = NULL;
    struct sockaddr_in myaddr;
    char *t;
    String s;


    bzero(&myaddr, sizeof(myaddr));
    if (fd < MAX_HTCP_SOCK) {
      myaddr = htcp_fdinfos[fd];
    }


    if (hdr->F1 == 1) {
	debug(31, 1) ("htcpHandleTstResponse: error condition, F1/MO == 1\n");

	connLogRecvICP(connID_none, from->sin_addr, from->sin_port,
		       myaddr.sin_addr, myaddr.sin_port,
		       ICP_ERR, servid, hdr->msg_id,
		       htcpReply.flags, 0);
	return;
    }
    memset(&htcpReply, '\0', sizeof(htcpReply));
    httpHeaderInit(&htcpReply.hdr, hoHtcpReply);
    htcpReply.msg_id = hdr->msg_id;
    if (hdr->response == 15) {
      htcpReply.flags |= ICP_FLAG_SETUP;
    }
    debug(31, 3) ("htcpHandleTstResponse: msg_id = %d\n", (int) htcpReply.msg_id);
    htcpReply.shostid = servid;
    htcpReply.hit = hdr->response ? 0 : 1;
    if (hdr->F1) {
	debug(31, 3) ("htcpHandleTstResponse: MISS\n");
    } else {
	debug(31, 3) ("htcpHandleTstResponse: HIT\n");
	d = htcpUnpackDetail(buf, sz);
	if (d == NULL) {
	    debug(31, 1) ("htcpHandleTstResponse: bad DETAIL\n");
	    connLogRecvICP(connID_none, from->sin_addr, from->sin_port,
			   myaddr.sin_addr, myaddr.sin_port,
			   ICP_ERR, servid, hdr->msg_id,
			   htcpReply.flags, 0);
	    return;
	}
	if ((t = d->resp_hdrs))
	    httpHeaderParse(&htcpReply.hdr, t, t + strlen(t));
	if ((t = d->entity_hdrs))
	    httpHeaderParse(&htcpReply.hdr, t, t + strlen(t));
	if ((t = d->cache_hdrs))
	    httpHeaderParse(&htcpReply.hdr, t, t + strlen(t));
    }

    s = httpHeaderGetByName(&htcpReply.hdr, "Cache-RTT");
    debug(31,3) ("Cache-RTT: %s\n", s.buf);
    if (s.buf != NULL) {
      htcpReply.cto.rtt = (double) atol(s.buf);
    }
    stringClean(&s);

  
    s = httpHeaderGetByName(&htcpReply.hdr, "Cached PCONN");
    debug(31,3) ("Cached PCONN: %s\n", s.buf);
    
    if (s.buf != NULL) {
      htcpReply.flags |= ICP_FLAG_SETUP_PCONN;
      connLogRecvICP(connID_none, from->sin_addr, from->sin_port,
		     myaddr.sin_addr, myaddr.sin_port,
		     ICP_MISS, servid, hdr->msg_id,
		     htcpReply.flags, htcpReply.cto.rtt);      
    }
    else {
      connLogRecvICP(connID_none, from->sin_addr, from->sin_port,
		     myaddr.sin_addr, myaddr.sin_port,
		     ICP_MISS, servid, hdr->msg_id,
		     htcpReply.flags, htcpReply.cto.rtt);      
    }
    stringClean(&s);


    key = queried_keys[htcpReply.msg_id % N_QUERIED_KEYS];
    debug(31, 3) ("htcpHandleTstResponse: key (%p) %s\n", key, storeKeyText(key));
    neighborsHtcpReply(key, &htcpReply, from);
    httpHeaderClean(&htcpReply.hdr);
    if (d)
	htcpFreeDetail(d);
}

static void
htcpHandleTstRequest(int fd, htcpDataHeader * dhdr, char *buf, int sz, struct sockaddr_in *from, int servid)
{
    /* buf should be a SPECIFIER */
    htcpSpecifier *s;
    StoreEntry *e;
    struct sockaddr_in myaddr;
    conn_id_t ConnID;

    bzero(&myaddr, sizeof(myaddr));
    if (fd < MAX_HTCP_SOCK) {
      myaddr = htcp_fdinfos[fd];
    }

    if (sz == 0) {
	debug(31, 3) ("htcpHandleTst: nothing to do\n");
	return;
    }
    if (dhdr->F1 == 0)
	return;
    s = htcpUnpackSpecifier(buf, sz);
    if (NULL == s) {
	debug(31, 3) ("htcpHandleTstRequest: htcpUnpackSpecifier failed\n");
	return;
    }
    debug(31, 3) ("htcpHandleTstRequest: %s %s %s\n",
	s->method,
	s->uri,
	s->version);
    debug(31, 3) ("htcpHandleTstRequest: %s\n", s->req_hdrs);
    ConnID = connLogNewICPReq(s->uri, from->sin_addr, from->sin_port,
			      myaddr.sin_addr, myaddr.sin_port,
			      servid, dhdr->msg_id,
			      dhdr->response ? 0 : 1);          // probably also response
    if ((e = htcpCheckHit(s)))
	htcpTstReply(fd, dhdr, e, s, from, ConnID);		/* hit */
    else if (dhdr->response == 15)                   // This is based on hdr.response
        icpHandleSpeculative(fd, from, 0,s->uri, servid, dhdr->msg_id, ConnID, PROTO_HTCP);
    else
	htcpTstReply(fd, dhdr, NULL, NULL, from, ConnID);	/* cache miss */
    htcpFreeSpecifier(s);
}

static void
htcpHandleMon(htcpDataHeader * hdr, char *buf, int sz, struct sockaddr_in *from)
{
    debug(31, 3) ("htcpHandleMon: Unimplemented\n");
}

static void
htcpHandleSet(htcpDataHeader * hdr, char *buf, int sz, struct sockaddr_in *from)
{
    debug(31, 3) ("htcpHandleSet: Unimplemented\n");
}

static void
htcpHandleData(int fd, char *buf, int sz, struct sockaddr_in *from)
{
    htcpDataHeader hdr;
    htcpAuthHeader auth;
    char *servid;
    int keynamesize;
    unsigned char *signature;
    static char authbuf[8192];
    int authoff = 0;
    int i, s;
    int server_id = 0;
    const cache_key *digest;

    if (sz < sizeof(htcpDataHeader)) {
	debug(31, 0) ("htcpHandleData: msg size less than htcpDataHeader size\n");
	return;
    }
    xmemcpy(&hdr, buf, sizeof(htcpDataHeader));
    hdr.length = ntohs(hdr.length);
    hdr.msg_id = ntohl(hdr.msg_id);
    debug(31, 3) ("htcpHandleData: sz = %d\n", sz);
    debug(31, 3) ("htcpHandleData: length = %d\n", (int) hdr.length);
    if (hdr.opcode >= HTCP_END) {
	debug(31, 0) ("htcpHandleData: client %s, opcode %d out of range\n",
	    inet_ntoa(from->sin_addr),
	    (int) hdr.opcode);
	return;
    }
    debug(31, 3) ("htcpHandleData: opcode = %d %s\n",
	(int) hdr.opcode, htcpOpcodeStr[hdr.opcode]);
    debug(31, 3) ("htcpHandleData: response = %d\n", (int) hdr.response);
    debug(31, 3) ("htcpHandleData: F1 = %d\n", (int) hdr.F1);
    debug(31, 3) ("htcpHandleData: RR = %d\n", (int) hdr.RR);
    debug(31, 3) ("htcpHandleData: msg_id = %d\n", (int) hdr.msg_id);
    if (sz < hdr.length) {
	debug(31, 0) ("htcpHandle: sz < hdr.length\n");
	return;
    }

    /* Basic idea: Same as before...keep an authbuf that we copy into.
     * Grab the src and dest addrs from function arguments
     * Make a htcpUnpackAuth function to build an auth struct.  use for sig_time and keyname
     * Just copy OP data straight from buf (given buf and hdr.length)
     * UGH
     */


    // First, don't even bother checking the auth if we have no secret key installed
    
    if (Config.htcpkey) {

      if (sz < hdr.length + sizeof(htcpAuthHeader)){
	debug(31,0) ("htcpHandle: sz < auth.length\n");
	return;
      }
      
      xmemcpy(&auth, buf + hdr.length, sizeof(htcpAuthHeader));
      
      keynamesize = htcpUnpackCountstr(buf + hdr.length + sizeof(htcpAuthHeader), 
				       sz - hdr.length - sizeof(htcpAuthHeader),
				       &servid);
      
      // TODO: Add in a check to make sure htcpUnpackSig exited with no errors
      htcpUnpackSig(buf + hdr.length + sizeof(htcpAuthHeader) + keynamesize, 
		    sz - hdr.length - sizeof(htcpAuthHeader) - keynamesize,
		    &signature);
      
      debug(31,3) ("htcpHandle: keyname = %s\n", servid);
      //debug(31,3) ("htcpHandle: signature = %s\n", signature);    
      
      // Build the authbuf;
      
      memset(&authbuf, '\0', 8192);
      authoff += 14;
      
      xmemcpy(authbuf + authoff, &auth.sig_time, 4);
      authoff += 4;
      xmemcpy(authbuf + authoff, &auth.sig_expire, 4);
      authoff += 4;
      
      auth.length = ntohs(auth.length);
      auth.sig_time = ntohl(auth.sig_time);
      auth.sig_expire = ntohl(auth.sig_expire);
      
      
      // Here we check to see if the auth has expired
      
      // Next comes the HTCP OP data
      xmemcpy(authbuf + authoff, buf, hdr.length);
      authoff += hdr.length;
      
      // Then the keyname
      s = htcpBuildCountstr(authbuf + authoff, 8192 - authoff, servid);
      if (s < 0)
	return;
      authoff += s;
      debug(31,3) ("htcpHandleData: authoff = %d\n", authoff);
      
      digest = hmac_md5(Config.htcpkey, strlen(Config.htcpkey), authbuf, authoff);
      
      // Last is the signature
      //s = htcpBuildCountstr(authbuf + authoff, 8192 - authoff, signature);
      //if (s < 0)
      //	return;
      //authoff += s;
      
      // Here we check the signature.
      
      if (memcmp(digest, signature, MD5_DIGEST_CHARS) != 0) {
	debug(31,0) ("htcpHandleData: AUTHENTICATION FAILED.  Dropping packet silently\n");
	for(i=0;i<MD5_DIGEST_CHARS;i++) {
	  fprintf(stderr, "%d:%X ", i, digest[i]);
	}
	fprintf(stderr, "\n");
	for(i=0;i<MD5_DIGEST_CHARS;i++) {
	  fprintf(stderr, "%d:%X ", i, signature[i]);
	}
	fprintf(stderr, "\n");
	return;
      }
      
      server_id = atol(servid);
      debug(31,3) ("ServerID: %d\n", server_id);


      // Gotta free the countstr
      safe_free(servid);
      safe_free(signature);

      // If the auth failed, then we exited the function.  Otherwise, we're all set
      // RRR: Is there a memory leak in here somewhere?
    }

    /*
     * set sz = hdr.length so we ignore any AUTH fields following
     * the DATA.
     */
    sz = (int) hdr.length;
    buf += sizeof(htcpDataHeader);
    sz -= sizeof(htcpDataHeader);
    debug(31, 3) ("htcpHandleData: sz = %d\n", sz);
    htcpHexdump("htcpHandleData", buf, sz);
    switch (hdr.opcode) {
    case HTCP_NOP:
	htcpHandleNop(&hdr, buf, sz, from);
	break;
    case HTCP_TST:
	htcpHandleTst(fd, &hdr, buf, sz, from, server_id);
	break;
    case HTCP_MON:
	htcpHandleMon(&hdr, buf, sz, from);
	break;
    case HTCP_SET:
	htcpHandleSet(&hdr, buf, sz, from);
	break;
    case HTCP_CLR:
	debug(31, 1) ("htcpHandleData: client %s, CLR not supported\n",
	    inet_ntoa(from->sin_addr));
	break;
    default:
	assert(0);
	break;
    }
}

static void
htcpHandle(int fd, char *buf, int sz, struct sockaddr_in *from)
{
    htcpHeader htcpHdr;
    if (sz < sizeof(htcpHeader)) {
	debug(31, 0) ("htcpHandle: msg size less than htcpHeader size\n");
	return;
    }
    htcpHexdump("htcpHandle", buf, sz);
    xmemcpy(&htcpHdr, buf, sizeof(htcpHeader));
    htcpHdr.length = ntohs(htcpHdr.length);
    debug(31, 3) ("htcpHandle: htcpHdr.length = %d\n", (int) htcpHdr.length);
    debug(31, 3) ("htcpHandle: htcpHdr.major = %d\n", (int) htcpHdr.major);
    debug(31, 3) ("htcpHandle: htcpHdr.minor = %d\n", (int) htcpHdr.minor);
    if (sz != htcpHdr.length) {
	debug(31, 1) ("htcpHandle: sz/%d != htcpHdr.length/%d from %s:%d\n",
	    sz, htcpHdr.length,
	    inet_ntoa(from->sin_addr), (int) ntohs(from->sin_port));
	return;
    }
    buf += sizeof(htcpHeader);
    sz -= sizeof(htcpHeader);
    htcpHandleData(fd, buf, sz, from);
}

static void
htcpRecv(int fd, void *data)
{
    static char buf[8192];
    int len;
    static struct sockaddr_in from;
    int flen = sizeof(struct sockaddr_in);
    memset(&from, '\0', flen);
    statCounter.syscalls.sock.recvfroms++;
    len = recvfrom(fd, buf, 8192, 0, (struct sockaddr *) &from, &flen);
    debug(31, 3) ("htcpRecv: FD %d, %d bytes from %s:%d\n",
	fd, len, inet_ntoa(from.sin_addr), ntohs(from.sin_port));
    htcpHandle(fd, buf, len, &from);
    commSetSelect(fd, COMM_SELECT_READ, htcpRecv, NULL, 0);
}

/*
 * ======================================================================
 * PUBLIC FUNCTIONS
 * ======================================================================
 */

void
htcpInit(void)
{
    sockaddr_in_list *sl;
    int count = 0;

    if (Config.Port.htcp <= 0) {
	debug(31, 1) ("HTCP Disabled.\n");
	return;
    }
    enter_suid();
    htcpInSocket = comm_open(SOCK_DGRAM,
	0,
	Config.Addrs.udp_incoming,
	Config.Port.htcp,
	COMM_NONBLOCKING,
	"HTCP Socket");
    leave_suid();
    if (htcpInSocket < 0)
	fatal("Cannot open HTCP Socket");

    htcp_fdinfos[htcpInSocket].sin_addr = Config.Addrs.udp_incoming;
    htcp_fdinfos[htcpInSocket].sin_port = htons(Config.Port.htcp);

    commSetSelect(htcpInSocket, COMM_SELECT_READ, htcpRecv, NULL, 0);
    debug(31, 1) ("Accepting HTCP messages on port %d, FD %d.\n",
	(int) Config.Port.htcp, htcpInSocket);
    if (Config.Addrs.udp_outgoing.s_addr != no_addr.s_addr) {
	enter_suid();
	htcpOutSocket = comm_open(SOCK_DGRAM,
	    0,
	    Config.Addrs.udp_outgoing,
	    Config.Port.htcp,
	    COMM_NONBLOCKING,
	    "Outgoing HTCP Socket");
	leave_suid();
	if (htcpOutSocket < 0)
	    fatal("Cannot open Outgoing HTCP Socket");

	htcp_fdinfos[htcpOutSocket].sin_addr = Config.Addrs.udp_outgoing;
	htcp_fdinfos[htcpOutSocket].sin_port = htons(Config.Port.htcp);

	commSetSelect(htcpOutSocket, COMM_SELECT_READ, htcpRecv, NULL, 0);
	debug(31, 1) ("Outgoing HTCP messages on port %d, FD %d.\n",
	    (int) Config.Port.htcp, htcpOutSocket);
	fd_note(htcpInSocket, "Incoming HTCP socket");
    } else {
	htcpOutSocket = htcpInSocket;
	debug(31, 1) ("Outgoing HTCP messages on port %d, FD %d.\n",
	    (int) Config.Port.htcp, htcpOutSocket);
    }
    if (!htcpSpecifierPool) {
	htcpSpecifierPool = memPoolCreate("htcpSpecifier", sizeof(htcpSpecifier));
	htcpDetailPool = memPoolCreate("htcpDetail", sizeof(htcpDetail));
    }

#if 1

    // TODO: Add in the analogue of all_addrs to hack in a failaddr list for htcp
    // TODO DONE: htcp_port lets us accept incoming messages on 0.0.0.0:4287
    // TODO DONE: htcp_addrs lets us bind to specific outgoing htcp ports

    debug(31,0) ("htcpInit: secret key = %s\n", Config.htcpkey ? "INSTALLED" : "NULL");

    if (ICPSenderID == 0) {
      ICPSenderID = squid_random();
      if ((HTCPSenderID = xmalloc(32)) < 0) {
	fatal("Can't allocate SenderID!");
      }
      snprintf(HTCPSenderID, 32, "%ld", ICPSenderID);
      debug(31,0) ("Initializing SenderID to: %s\n", HTCPSenderID);
    }


    for (sl = Config.Sockaddr.udp;
         sl && count < MAX_HTCP_ADDRS;
         sl = sl->next, count++) {
        int fd;
        enter_suid();
        fd = comm_open(SOCK_DGRAM,
                       0,
                       sl->s.sin_addr,
                       Config.Port.htcp,
                       COMM_NONBLOCKING,
                       "HTCP Specific Port");
        leave_suid();
        if (fd < 0) {
            debug(12, 1)("Cannot open Outgoing HTCP Port for '%s' %d: %s",
                  inet_ntoa(sl->s.sin_addr), Config.Port.htcp, xstrerror());
            count--;
            continue;
        }
        commSetSelect(fd,
            COMM_SELECT_READ,
		      htcpRecv, //icpHandleUdp,     //********** TODO: CHANGE ME!
            NULL,
            0);
        debug(12, 1) ("Sub Outgoing HTCP messages on %s:%d, FD %d.\n",
                      inet_ntoa(sl->s.sin_addr), (int) Config.Port.htcp,
                      fd);
        fd_note(fd, "Outgoing HTCP Sub socket");
        fd_note(fd, "Incoming HTCP Sub socket");
        htcp_fdaddrs[count].sa = sl->s;
        htcp_fdaddrs[count].fd = fd;
        n_htcp_fdaddrs++;
        if (fd < MAX_HTCP_SOCK) {
            htcp_fdinfos[fd].sin_addr = sl->s.sin_addr;
            htcp_fdinfos[fd].sin_port = htons(Config.Port.htcp);
        }
    }

#endif

}

void
htcpQuery(StoreEntry * e, request_t * req, peer * p, conn_id_t reqid, int reqnum)
{
    cache_key *save_key;
    char *pkt;
    ssize_t pktlen;
    char vbuf[32];
    htcpStuff stuff;
    HttpHeader hdr;
    Packer pa;
    MemBuf mb;
    http_state_flags flags;

    if (htcpInSocket < 0)
	return;

    memset(&flags, '\0', sizeof(flags));
    snprintf(vbuf, sizeof(vbuf), "%d/%d",
	req->http_ver.major, req->http_ver.minor);
    stuff.op = HTCP_TST;
    stuff.rr = RR_REQUEST;
    stuff.f1 = 1;
    stuff.response = p->options.monet ? 15 : 0;
    debug(31, 3) ("htcpQuery: MONET Mode = %d, response = %d\n", p->options.monet, stuff.response);
    //stuff.msg_id = ++msg_id_counter;
    stuff.msg_id = reqnum;
    stuff.S.method = (char *) RequestMethodStr[req->method];
    stuff.S.uri = (char *) storeUrl(e);
    stuff.S.version = vbuf;
    httpBuildRequestHeader(req, req, e, &hdr, flags);
    memBufDefInit(&mb);
    packerToMemInit(&pa, &mb);
    httpHeaderPackInto(&hdr, &pa);
    httpHeaderClean(&hdr);
    packerClean(&pa);
    stuff.S.req_hdrs = mb.buf;

    /* Rohit: I've removed the src/dst fields from the authentication,
     * so now we can use the same packet for everything.  I have no
     * plans to change this, so i'm going to implement htcpMultiSend, 
     * which will send the same packet out from each interface.  
     * Therefore, I only need to call htcpMyAddrListSend() from here.
     */

    if (p->options.monet) {
      pkt = htcpBuildMonetPacket(&stuff, &pktlen);
    }
    else {
      pkt = htcpBuildPacket(&stuff, &pktlen);
    }

    if (pkt == NULL) {
      debug(31, 0) ("htcpQuery: HTCP Packet Build failed\n");
      return;
    }  

    // What happens here if I am not talking to a monet node?
    htcpMyAddrlistSend(pkt, (int) pktlen, &p->in_addr, reqid, &stuff);

    /* Common cleanup here */
    memBufClean(&mb);
    save_key = queried_keys[stuff.msg_id % N_QUERIED_KEYS];
    storeKeyCopy(save_key, e->hash.key);
    debug(31, 3) ("htcpQuery: key (%p) %s\n", save_key, storeKeyText(save_key));

    xfree(pkt);
}

/*  
 * htcpSocketShutdown only closes the 'in' socket if it is
 * different than the 'out' socket.
 */
void
htcpSocketShutdown(void)
{
    if (htcpInSocket < 0)
	return;
    if (htcpInSocket != htcpOutSocket) {
	debug(12, 1) ("FD %d Closing HTCP socket\n", htcpInSocket);
	comm_close(htcpInSocket);
    }
    /*      
     * Here we set 'htcpInSocket' to -1 even though the HTCP 'in'
     * and 'out' sockets might be just one FD.  This prevents this
     * function from executing repeatedly.  When we are really ready to
     * exit or restart, main will comm_close the 'out' descriptor.
     */
    htcpInSocket = -1;
    /*      
     * Normally we only write to the outgoing HTCP socket, but
     * we also have a read handler there to catch messages sent
     * to that specific interface.  During shutdown, we must
     * disable reading on the outgoing socket.
     */
    /* XXX Don't we need this handler to read replies while shutting down?
     * I think there should be a separate hander for reading replies..
     */
    assert(htcpOutSocket > -1);
    commSetSelect(htcpOutSocket, COMM_SELECT_READ, NULL, NULL, 0);

    // TODO DONE: close all the incoming failaddrs

    if (n_htcp_fdaddrs > 0) {
	int i;
	for (i = 0; i < n_htcp_fdaddrs; i++) {
	    debug(12, 1) ("FD %d Closing HTCP Sub socket\n", htcp_fdaddrs[i].fd);
	    comm_close(htcp_fdaddrs[i].fd);
	    htcp_fdaddrs[i].fd = -1;
	}
	n_htcp_fdaddrs = 0;
    }    

}

void
htcpSocketClose(void)
{
    htcpSocketShutdown();
    if (htcpOutSocket > -1) {
	debug(12, 1) ("FD %d Closing HTCP socket\n", htcpOutSocket);
	comm_close(htcpOutSocket);
	htcpOutSocket = -1;
    }
}


/* Rohit: Called by specConnectDone() in icp_v2.c
 * Builds the appropriate htcp packet (PCONN, Success, or error)
 * We're going to keep the buggy squid HTCP style and always put in the htcpDetail
 * If we have an error, set f1 = 1
 * If we have a pconn, set a PCONN cache header
 * Also, add in a Cache-to-origin: header
 */
void
htcpTstMonetReply(u_num32 msg_id, int n_to_send, struct sockaddr_in *from, int *to, 
		  conn_id_t ConnID, int rtt, int status)
{
    htcpStuff stuff;
    char *pkt;
    HttpHeader hdr;
    MemBuf mb;
    Packer p;
    ssize_t pktlen;
    char *host;
    char cto_buf[128];
    int n_sent = 0;
    int i;

    memset(&stuff, '\0', sizeof(stuff));
    stuff.op = HTCP_TST;
    stuff.rr = RR_RESPONSE;

    // Need to vary this by input...
    if (status == HTCP_STATUS_ERROR) {
      stuff.f1 = 1;
    }
    else {
      stuff.f1 = 0;
    }

    // Response is always 15 (generic MONET flag)
    stuff.response = 15;
    stuff.msg_id = msg_id;

    // Fill in our Cache Headers
    memBufDefInit(&mb);
    packerToMemInit(&p, &mb);
    httpHeaderInit(&hdr, hoHtcpReply);
    snprintf(cto_buf, 128, "%d", rtt);
    httpHeaderPutExt(&hdr, "Cache-RTT", cto_buf);
    
    if (status == HTCP_STATUS_PCONN) {
      httpHeaderPutExt(&hdr, "Cached PCONN", "1");
    }

    httpHeaderPackInto(&hdr, &p);
    stuff.D.cache_hdrs = xstrdup(mb.buf);
    debug(31, 3) ("htcpTstMonetReply: cache_hdrs = {%s}\n", stuff.D.cache_hdrs);
    memBufClean(&mb);
    httpHeaderClean(&hdr);
    packerClean(&p);

    pkt = htcpBuildMonetPacket(&stuff, &pktlen);
    if (pkt == NULL) {
	debug(31, 0) ("htcpTstMonetReply: htcpBuildPacket() failed\n");
	return;
    }

    htcpMultiSend(pkt, (int) pktlen, n_to_send, to, from, ConnID, &stuff);

    xfree(pkt);
}
